from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Resource


@Resource("s3")
class DuploS3(DuploTenantResourceV3):
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, "aws/s3bucket")

  def name_from_body(self, body):
    return body["Name"]


import glob
import os
import time
from duplocloud.client import DuploClient
from duplocloud.errors import DuploError
from duplocloud.resource import DuploResource
from duplocloud.commander import Resource, Command
import duplocloud.args as args
import boto3
from multiprocessing.pool import ThreadPool 

@Resource("aws")
class DuploAWS(DuploResource):
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
    self.__creds = None
    self.__region = None
    self.__tenant_svc = duplo.load("tenant")

  # def __call__(self, *args):
  #   return {
  #     "message": "Hello from AWS plugin dude!"
  #   }
  
  def client(self, 
             name: str, 
             region: str = None, 
             refresh: bool = False):
    if not self.__creds or refresh:
      jit_svc = self.duplo.load("jit")
      self.__creds = jit_svc.aws()
      # if admin then we need the correct region from the tenant
      if self.duplo.tenant and self.duplo.isadmin:
        r = self.__tenant_svc.region()
        self.__creds["Region"] = r["region"]
    c = self.__creds
    return boto3.client(
        name,
        aws_access_key_id=c.get('AccessKeyId'),
        aws_secret_access_key=c.get('SecretAccessKey'),
        aws_session_token=c.get('SessionToken'),
        region_name=region or c.get('Region')
    )
  
  @Command()
  def update_website(self, 
                     name: args.NAME,
                     dir: args.CONTENT_DIR="dist",
                     wait: args.WAIT=False):
    """Update website
    
    Updates a static website hosted on S3 bucket and served via CloudFront. 
    First the contents of the bucket are updated and then the CloudFront cache is invalidated.

    """
    # make sure the dir directory exists
    if not os.path.exists(dir):
      raise DuploError(f"Directory '{dir}' does not exist.", 404)

    # get the needed clients
    s3 = self.client("s3")
    cdf = self.client("cloudfront")

    # scope into tenant
    tenant = self.__tenant_svc.find()
    tenant_name = tenant["AccountName"]

    # find the distribution
    distributions = cdf.list_distributions()
    comment = f"duploservices-{tenant_name}-{name}"
    dist = next((
      d for d in distributions["DistributionList"]["Items"] 
      if d["Comment"] == comment
    ), None)
    if not dist:
      raise DuploError(f"CloudFront distribution '{comment}' not found.", 404)
    dist_id = dist["Id"]

    # find the bucket name from the distribution
    s3_domain = next((
      o["DomainName"] for o in dist["Origins"]["Items"] 
      if o["Id"] == dist["DefaultCacheBehavior"]["TargetOriginId"]
    ), None)
    bucket_name = s3_domain.split(".")[0]
    
    # get current list of objects in the bucket
    live_objs = s3.list_objects_v2(Bucket=bucket_name)
    live_files = [o["Key"] for o in live_objs.get("Contents", [])]

    # get flat list of files in dir
    tree = glob.glob('**/*', root_dir=dir,recursive=True, include_hidden=True)
    files = []
    # add a trailing slash to each directory because s3 folders have a slash :/
    for n, p in enumerate(tree):
      if os.path.isdir(f"{dir}/{p}"):
        tree[n] = f"{p}/"
      else: # collect the files
        files.append(p)

    # delete objects not in the local tree
    del_objs = [{"Key": o} for o in list(set(live_files) - set(tree))]
    if del_objs:
      s3.delete_objects(Bucket=bucket_name, Delete={"Objects": del_objs})
      for o in del_objs:
        self.duplo.logger.info(f"Deleted {o['Key']}")

    # upload the contents of the dir to the bucket
    pool = ThreadPool(10)
    pool.map(
      lambda key: s3.upload_file(
        Filename=f"{dir}/{key}", 
        Bucket=bucket_name, 
        Key=key,
        Callback=lambda x: self.duplo.logger.info(f"Uploading {x}B to {key}")
      ), 
      files
    )

    # invalidate the CloudFront cache
    call_ref = str(time.time()).replace(".", "")
    inv = cdf.create_invalidation(
      DistributionId=dist_id,
      InvalidationBatch={
        "Paths": {
          "Quantity": 1,
          "Items": ["/*"]
        },
        "CallerReference": call_ref
      }
    )

    # now wait for the invalidation to complete
    if wait:
      waiter = cdf.get_waiter('invalidation_completed')
      waiter.config.delay = 10
      waiter.config.max_attempts = 60
      waiter.wait(
        DistributionId=dist_id, 
        Id=inv['Invalidation']['Id']
      )
      self.duplo.logger.info(f"Invalidation by {call_ref} completed.")

    return {
      "message": f"Updated website '{name}'",
      "distribution": dist_id,
      "bucket": bucket_name,
      "pruned": len(del_objs),
      "uploaded": len(files)
    }




from . import args
from .client import DuploClient
from .errors import DuploError, DuploFailedResource
from .commander import get_parser, extract_args, Command
import math
import time

class DuploCommand():
  def __init__(self, duplo: DuploClient):
    self.duplo = duplo
  
  def __call__(self, *args):
    pass

class DuploResource():

  def __init__(self, duplo: DuploClient):
    self.duplo = duplo
    self.__logger = None
  
  def __call__(self, cmd: str, *args):
    c = self.command(cmd)
    return c(*args)

  # TODO: something is off and the logs will duplicate if we do this. Plese figure out how to actually create a logger for each resource.
  # @property
  # def logger(self):
  #   if not self.__logger:
  #     self.__logger = self.duplo.logger_for(self.__class__.__name__)
  #   return self.__logger
  
  def command(self, name: str):
    # TODO: Test the aliased_method further before actually releasing this feature. This will be disabled for now.
    # method = aliased_method(self.__class__, name)
    if not (command := getattr(self, name, None)):
      raise DuploError(f"Invalid command: {name}")
    cliargs = extract_args(command)
    parser = get_parser(cliargs)
    def wrapped(*args):
      pargs = vars(parser.parse_args(args))
      return command(**pargs)
    return wrapped
  
  def wait(self, wait_check: callable, timeout: int=None, poll: int=10):
    timeout = timeout or self.duplo.timeout
    exp = math.ceil(timeout / poll)
    for _ in range(exp):
      try:
        wait_check()
        break
      except DuploFailedResource as e:
        raise e
      except DuploError as e:
        if e.message:
          self.duplo.logger.debug(e)
        time.sleep(poll)
      except KeyboardInterrupt as e:
        raise e
    else:
      raise DuploError("Timed out waiting", 404)
    
class DuploResourceV2(DuploResource):

  def name_from_body(self, body):
    return body["Name"]
  def endpoint(self, path: str=None):
    return path
  @Command()
  def list(self) -> list:
    """Retrieve a List of {{kind}}
    
    Usage: cli usage
      ```sh
      duploctl {{kind | lower}} list
      ```

    Returns:
      list: A list of {{kind}}.
    """
    response = self.duplo.get(self.endpoint(self.paths["list"]))
    return response.json()
  @Command()
  def find(self, 
           name: args.NAME) -> dict:
    """Find a {{kind}} by name.

    Usage: cli usage
      ```sh
      duploctl {{kind | lower}} find <name>
      ```
    
    Args:
      name: The name of the {{kind}} to find.

    Returns: 
      resource: The {{kind}} object.
      
    Raises:
      DuploError: If the {{kind}} could not be found.
    """
    try:
      return [s for s in self.list() if self.name_from_body(s) == name][0]
    except IndexError:
      raise DuploError(f"{self.kind} '{name}' not found", 404)
      
  @Command()
  def apply(self,
             body: args.BODY,
             wait: args.WAIT = False):
    """Apply a service."""
    name = self.name_from_body(body)
    try:
      self.find(name)
      return self.update(name, body, wait)
    except DuploError:
      return self.create(body, wait)
  
class DuploTenantResourceV2(DuploResourceV2):
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
    self.__tenant = None
    self.__tenant_id = None
    self.tenant_svc = duplo.load('tenant')
  @property
  def tenant(self):
    if not self.__tenant:
      self.__tenant = self.tenant_svc.find()
      self.__tenant_id = self.__tenant["TenantId"]
    return self.__tenant
  
  @property
  def tenant_id(self):
    if not self.__tenant_id:
      if self.__tenant:
        self.__tenant_id = self.__tenant["TenantId"]
      elif self.duplo.tenantid:
        self.__tenant_id = self.duplo.tenantid
      else:
        self.__tenant_id = self.tenant["TenantId"]
    return self.__tenant_id
  
  def endpoint(self, path: str=None):
    return f"subscriptions/{self.tenant_id}/{path}"
  

class DuploTenantResourceV3(DuploResource):
  def __init__(self, duplo: DuploClient, slug: str):
    super().__init__(duplo)
    self.__tenant = None
    self.__tenant_id = None
    self.tenant_svc = duplo.load('tenant')
    self.slug = slug
    self.wait_timeout = 200
    self.wait_poll = 10
  @property
  def tenant(self):
    if not self.__tenant:
      self.__tenant = self.tenant_svc.find()
      self.__tenant_id = self.__tenant["TenantId"]
    return self.__tenant
  
  @property
  def tenant_id(self):
    if not self.__tenant_id:
      if self.__tenant:
        self.__tenant_id = self.__tenant["TenantId"]
      elif self.duplo.tenantid:
        self.__tenant_id = self.duplo.tenantid
      else:
        self.__tenant_id = self.tenant["TenantId"]
    return self.__tenant_id
  
  @Command()
  def list(self) -> list:
    """Retrieve a List of {{kind}} resources

    Usage: cli usage
      ```sh
      duploctl {{kind | lower}} list
      ```
    
    Returns:
      list: A list of {{kind}}.
    """
    response = self.duplo.get(self.endpoint())
    return response.json()
  
  @Command()
  def find(self, 
           name: args.NAME) -> dict:
    """Find {{kind}} resources by name.

    Usage: cli usage
      ```sh
      duploctl {{kind | lower}} find <name>
      ```
    
    Args:
      name: The name of the {{kind}} resource to find.

    Returns: 
      resource: The {{kind}} object.
      
    Raises:
      DuploError: If the {{kind}} could not be found.
    """
    response = self.duplo.get(self.endpoint(name))
    return response.json()
  
  @Command()
  def delete(self, 
             name: args.NAME) -> dict:
    """Delete a {{kind}} resource by name.

    Usage: cli usage
      ```sh
      duploctl {{kind | lower}} delete <name>
      ```
    
    Args:
      name: The name of the {{kind}} resource to delete.

    Returns: 
      message: A success message.

    Raises:
      DuploError: If the {{kind}} resource could not be found or deleted. 
    """
    self.duplo.delete(self.endpoint(name))
    return {
      "message": f"{self.slug}/{name} deleted"
    }
  
  @Command()
  def create(self, 
             body: args.BODY,
             wait: args.WAIT=False,
             wait_check: callable=None) -> dict:
    """Create a {{kind}} resource.

    Usage: CLI Usage
      ```sh
      duploctl {{kind | lower}} create -f '{{kind | lower}}.yaml'
      ```
      Contents of the `{{kind|lower}}.yaml` file
      ```yaml
      --8<-- "src/tests/data/{{kind|lower}}.yaml"
      ```

    Example: One liner example
      ```sh
      echo \"\"\"
      --8<-- "src/tests/data/{{kind|lower}}.yaml"
      \"\"\" | duploctl {{kind | lower}} create -f -
      ```
    
    Args:
      body: The resource to create.
      wait: Wait for the resource to be created.
      wait_check: A callable function to check if the resource

    Returns: 
      message: Success message.

    Raises:
      DuploError: If the resource could not be created.
    """
    name = self.name_from_body(body)
    response = self.duplo.post(self.endpoint(), body)
    if wait:
      self.wait(wait_check or (lambda: self.find(name)), self.wait_timeout, self.wait_poll)
    return response.json()
  
  @Command()
  def update(self, 
             body: args.BODY):
    """Update a V3 resource by name.
    
    Args:
      body (str): The resource to update.
    Returns: 
      Success message.
    Raises:
      DuploError: If the resource could not be created.
    """
    name = self.name_from_body(body)
    response = self.duplo.put(self.endpoint(name), body)
    return response.json()
  
  @Command()
  def apply(self,
             body: args.BODY,
             wait: args.WAIT = False):
    """Apply a service."""
    name = self.name_from_body(body)
    try:
      self.find(name)
      return self.update(name, body, wait)
    except DuploError:
      return self.create(body, wait)
  
  def name_from_body(self, body):
    return body["metadata"]["name"]

  def endpoint(self, name: str=None, path: str=None):
    p = f"v3/subscriptions/{self.tenant_id}/{self.slug}"
    if name:
      p += f"/{name}"
    if path:
      p += f"/{path}"
    return p



import time
from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV2
from duplocloud.errors import DuploError, DuploFailedResource
from duplocloud.commander import Command, Resource
from json import dumps, loads
import duplocloud.args as args

_STATUS_CODES = {
  "1": "Running",
  "3": "Pending",
  "4": "Waiting",
  "6": "Deleted",
  "7": "Failed",
  "11": "Succeeded"
}

@Resource("service")
class DuploService(DuploTenantResourceV2):
  """Duplocloud Service Resource
  
  This resource is used to manage services in Duplocloud. Using the `duploctl` command line tool, you can manage services with actions:
  
  Usage: Basic CLI Use  
    ```sh
    duploctl service <action>
    ```
  """
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
    self.paths = {
      "list": "GetReplicationControllers"
    }
    self.__pod_svc = self.duplo.load("pod")
    self.__old = None
    self.__updates = None
    self.__old_replicaset = None
    
  @Command()
  def update(self, 
             name: args.NAME,
             body: args.BODY = None,
             patches: args.PATCHES = None,
             wait: args.WAIT = False) -> dict:
    """Update a service.
    
    Update the state of a service.

    Usage: Basic CLI Use
      Update the replicas to 3 for a service.
      ```sh
      duploctl service update <service-name> --replace Replicas 3 
      ```
    
    Args:
      name: The name of the service to update.
      body: The body of the service to update.
      patches: A list of JSON patches as args to apply to the service.
        The options are `--add`, `--remove`, `--replace`, `--move`, and `--copy`.
        Then followed by `<path>` and `<value>` for `--add`, `--replace`, and `--test`.
      wait: Whether to wait for the service to update.
    
    """
    if (name is None and body is None):
      raise DuploError("No arguments provided to update service", 400)
    if name and body:
      body["Name"] = name
    if body is None:
      body = self.find(name)
    # before patching, save the original body if we need to wait
    old = None
    if wait:
      old = self.find(name) # should use the cache
      old["Replicaset"] = self.current_replicaset(name)
    if patches:
      body = self.duplo.jsonpatch(body, patches)
    if ((ttags := body["Template"].get("AllocationTags", None))
        and not body.get("AllocationTags", None)):
      body["AllocationTags"] = ttags
    body["OtherDockerConfig"] = body["Template"]["OtherDockerConfig"]
    self.duplo.post(self.endpoint("ReplicationControllerChangeAll"), body)
    if wait:
      self.wait(old, body)
    return {
      "message": f"Successfully updated service '{body['Name']}'"
    }
  
  @Command()
  def create(self,
             body: args.BODY,
             wait: args.WAIT = False) -> dict:
    """Create a service.
    
    Create a service in Duplocloud.
    
    Usage: Basic CLI Use
      ```sh
      duploctl service create --file service.yaml
      ```
      Contents of the `service.yaml` file
      ```yaml
      --8<-- "src/tests/data/service.yaml"
      ```

    Args:
      body: The service to create.
      wait: Wait for the service to be created.

    Returns:
      message: Success message.
    """
    self.duplo.post(self.endpoint("ReplicationControllerUpdate"), body)
    if wait:
      self.wait(lambda: self.find(body["Name"]))
    return {
      "message": f"Successfully created service '{body['Name']}'"
    }

  @Command()
  def delete(self,
             name: args.NAME) -> dict:
    """Delete a service.
    
    Delete a service in Duplocloud.

    Usage: Basic CLI Use
      ```sh
      duploctl service delete <service-name>
      ```

    Args:
      name: The name of the service to delete.

    Returns:
      message: Success message.
    """
    body = {
      "Name": name,
      "State": "delete"
    }
    self.duplo.post(self.endpoint("ReplicationControllerUpdate"), body)
    return {
      "message": f"Successfully deleted service '{name}'"
    }

  @Command()
  def update_replicas(self, 
                      name: args.NAME,
                      replica: args.REPLICAS,
                      wait: args.WAIT = False) -> dict:
    """Scale Service

    Update the number of replicas for a service.

    Usage: Basic CLI Use
      ```sh
      duploctl service update_replicas <service-name> <replicas>
      ```

    Args:
        name (str): The name of the service to update.
        replica (str): Number of replicas to set for service.
    """
    service = self.find(name)
    data = {
      "Name": name,
      "Replicas": replica,
      "AllocationTags": service["Template"].get("AllocationTags", "")
    }
    self.duplo.post(self.endpoint("ReplicationControllerChange"), data)
    if wait:
      self.wait(service, data)
    return {"message": f"Successfully updated replicas for service '{name}'"} 
  
  @Command()
  def update_image(self, 
                   name: args.NAME, 
                   image: args.IMAGE,
                   wait: args.WAIT = False) -> dict:
    """Update the image of a service.

    Update the image of a service.

    Usage: Basic CLI Use
      ```sh
      duploctl service update_image <service-name> <image-name>
      ```
    
    Args:
      name: The name of the service to update.
      image: The new image to use for the service.

    Returns:
      message: Success message
    """
    service = self.find(name)
    current_image =  self.image_from_body(service)
    data = {
      "Name": name,
      "Image": image,
      "AllocationTags": service["Template"].get("AllocationTags", "")
    }

    # needed before the update so we know how many pods to wait for
    if wait:
      service["Replicaset"] = self.current_replicaset(name)

    # if the tag is the same, reboot to pull new code, otherwise actual update
    if(current_image == image):
      self.duplo.post(self.endpoint(f"ReplicationControllerReboot/{name}"))
    else:
      self.duplo.post(self.endpoint("ReplicationControllerChange"), data)
      
    if wait:
      self.wait(service, data)

    return {"message": f"Successfully updated image for service '{name}'"}
 
  @Command()
  def update_env(self, 
                 name: args.NAME,
                 setvar: args.SETVAR,
                 strategy: args.STRATEGY,
                 deletevar: args.DELETEVAR,
                 wait: args.WAIT = False) -> dict:
    """Update the environment variables of a service. If service has no environment variables set, use -strat replace to set new values.
    Usage: Basic CLI Use
      ```sh
      duploctl service update_env <service-name> --setvar env-key1 env-val1 --setvar env-key2 env-val2 --setvar env-key3 env-val3 -strat merge --host $DUPLO_HOST --tenant $DUPLO_TENANT --token $DUPLO_TOKEN
      ```
    Args:
      name (str): The name of the service to update.

      setvar/-V (list): A list of key value pairs to set as environment variables.
      
      strategy/strat (str): The merge strategy to use for env vars. Valid options are "merge" or "replace".  Default is merge.

      deletevar/-D (list): A list of keys to delete from the environment variables.
    """
    # Returns an array of key and value mappings with provided keys
    def new_env_vars(setvar, key_name="Name", value_name="Value"):
      return [{key_name: i[0], value_name: i[1]} for i in setvar]
  
    def detect_case_format(env_list):
      # Check if the environment variables are using upper, lower, or mixed case for Name and Value
      if all('Name' in env and 'Value' in env for env in env_list):
          return "title"
      elif all('name' in env and 'value' in env for env in env_list):
          return "lower"
      # No consistent standard
      return "mixed"
    
    service = self.find(name)
    currentDockerconfig = loads(service["Template"]["OtherDockerConfig"])
    currentEnv = currentDockerconfig.get("Env", [])
    # Check if user is attempting to merge against a null Env. If so, set currentEnv to empty.
    if currentEnv is None and strategy == "merge":
      self.duplo.logger.warn("Specified \"merge\" strategy on a service with"
                             " no environment variables defined, should use "
                             " \"replace\". Proceeding anyway")
      currentEnv = []
    case_format = detect_case_format(currentEnv)
    if strategy == 'merge':
      try:
        if case_format == "title":
          newEnv = new_env_vars(setvar) if setvar is not None else []
          d = {env['Name']: env for env in currentEnv + newEnv}
        elif case_format == "lower":
          newEnv = new_env_vars(setvar, key_name="name", value_name="value") if setvar is not None else []
          d = {env['name']: env for env in currentEnv + newEnv}
        else:
          self.duplo.logger.warn("Possible attempt to merge env vars with"
                                 " inconsistent Name/Value key case."
                                 " Normalzing to capitalized"
                                 " \"Name\" and \"Value\"")
          norm_currentEnv = [{k.capitalize(): v for k, v in env.items()} for env in currentEnv]
          newEnv = new_env_vars(setvar,) if setvar is not None else []
          d = {env['Name']: env for env in norm_currentEnv + newEnv}
      except KeyError:
        raise DuploError("Could not merge new and existing environment variables")
      mergedvars = list(d.values())
      currentDockerconfig['Env'] = mergedvars
    else:
      newEnv = new_env_vars(setvar) if setvar is not None else []
      currentDockerconfig['Env'] = newEnv
    if deletevar is not None:
      for key in deletevar:
        try:
          currentDockerconfig['Env'] = [d for d in currentDockerconfig['Env'] if d['Name'] != key]
        except KeyError:
          currentDockerconfig['Env'] = [d for d in currentDockerconfig['Env'] if d['name'] != key] 
    payload = {
      "Name": name,
      "OtherDockerConfig": dumps(currentDockerconfig),
      "allocationTags": service["Template"].get("AllocationTags", "")
    }
    self.duplo.post(self.endpoint("ReplicationControllerChange"), payload)
    if wait:
      service["Replicaset"] = self.current_replicaset(name)
      self.wait(service, payload)
    return {"message": f"Successfully updated environment variables for service '{name}'"}
    
  @Command()
  def update_pod_label(self,
                 name: args.NAME,
                 setvar: args.SETVAR,
                 strategy: args.STRATEGY,
                 deletevar: args.DELETEVAR,
                 wait: args.WAIT = False):    
  
    """Update the pod labels of a service. If service has no pod labels set, use -strat replace to set new values.
    Usage: Basic CLI Use
      ```sh
      duploctl service update_pod_label <service-name> --setvar env-key1 env-val1 --setvar env-key2 env-val2 --setvar env-key3 env-val3 -strat merge --host $DUPLO_HOST --tenant $DUPLO_TENANT --token $DUPLO_TOKEN
      duploctl service update_pod_label <service-name> --setvar env-key1 env-val1 --setvar env-key2 env-val2 -strat replace --host $DUPLO_HOST --tenant $DUPLO_TENANT --token $DUPLO_TOKEN
      duploctl service update_pod_label <service-name> --deletevar env-key1 --host $DUPLO_HOST --tenant $DUPLO_TENANT --token $DUPLO_TOKEN
      
      ```
    Args:
      name: The name of the service to update.
      setvar/-V: A list of key value pairs to set as environment variables.
      strategy/strat: The merge strategy to use for env vars. Valid options are "merge" or "replace".  Default is merge.
      deletevar/-D: A list of keys to delete from the environment variables.
    """
    service = self.find(name)
    currentDockerconfig = loads(service["Template"]["OtherDockerConfig"])
    currentLabels = currentDockerconfig.get("PodLabels", [])
    newLabels = []
    if setvar is not None:
      newLabels = [{"Name": i[0], "Value": i[1]} for i in setvar]

    # merge new labels in existing labels (append)
    if strategy == 'merge':
      mergedLabels = currentLabels
      for vars in newLabels:
        mergedLabels[vars["Name"]] = vars["Value"]
      currentDockerconfig['PodLabels'] = mergedLabels

    else:
      # replace values of existing label
      for label, value in currentDockerconfig['PodLabels'].items():
        for vars in newLabels:
          if vars["Name"] not in currentDockerconfig['PodLabels']:
            label_name = vars["Name"]
            raise DuploError(f"Service {name} does not have Pod label '{label_name}', enter correct label.", 400)
          if vars["Name"] == label:
            currentDockerconfig['PodLabels'][label] = vars["Value"]

    if deletevar is not None:
      for key in deletevar:
        del currentDockerconfig['PodLabels'][key]
      
    payload = {
      "Name": name,
      "OtherDockerConfig": dumps(currentDockerconfig),
      "allocationTags": service["Template"].get("AllocationTags", "")
    }
    self.duplo.post(self.endpoint("ReplicationControllerChange"), payload)
    if wait:
      service["Replicaset"] = self.current_replicaset(name)
      self.wait(service, payload)
    return {"message": f"Successfully updated pod labels for service '{name}'"}
  
  @Command()
  def bulk_update_image(self, 
                        serviceimage: args.SERVICEIMAGE,
                        wait: args.WAIT = False) -> dict:
    """
    Bulk update the image of multiple services.

    Usage: Basic CLI Use
      ```sh
      duploctl service bulk_update_image -S <service-name-1> <image-name-1> -S <service-name-2> <image-name-2>
      ```

    Args:
      serviceimage: Takes n sets of two arguments, service name and image name. 
                    e.g., -S service1 image1:tag -S service2 image2:tag
      wait: Boolean flag to wait for service updates.
    """
    payload = []
    wait_list = []
    for name, image in serviceimage.items():
      service = self.find(name)
      payload_item = {
          "Name": name,
          "Image": image,
          "AllocationTags": service["Template"]["AllocationTags"]
      }
      payload.append(payload_item)
      if wait:
        wait_list.append({
          "old": service,
          "updated": payload_item
        })

    self.duplo.post(self.endpoint("ReplicationControllerBulkChangeAll"), payload)

    if wait:
      for update_info in wait_list:
        self.wait(**update_info)

    return {"message": "Successfully updated images for services"}

  @Command()
  def restart(self, 
              name: args.NAME,
              wait: args.WAIT = False) -> dict:
    """Restart a service.

    Restart a service.

    Usage: Basic CLI Use
      ```sh
      duploctl service restart <service-name>
      ```
    
    Args:
      name: The name of the service to restart.
      wait: Boolean flag to wait for service updates.

    Returns: 
      A success message if the service was restarted successfully.

    Raises:
      DuploError: If the service could not be restarted.
    """
    self.duplo.post(self.endpoint(f"ReplicationControllerReboot/{name}"))
    if wait:
      service = self.find(name)
      self.wait(service, service)
    return {"message": f"Successfully restarted service '{name}'"}
  
  @Command()
  def stop(self, 
           name: args.NAME,
           wait: args.WAIT = False) -> dict:
    """Stop a service.

    Stop a service.

    Usage: Basic CLI Use
      ```sh
      duploctl service stop <service-name>
      ```
    
    Args:
      name (str): The name of the service to stop.
      wait: Boolean flag to wait for service updates.

    Returns: 
      A success message if the service was stopped successfully.  
      
    Raises:
      DuploError: If the service could not be stopped.  
    """
    self.duplo.post(self.endpoint(f"ReplicationControllerStop/{name}"))
    if wait:
      service = self.find(name)
      self.wait(service, service)
    return {"message": f"Successfully stopped service '{name}'"}
  
  @Command()
  def start(self, 
            name: args.NAME,
            wait: args.WAIT = False) -> dict:
    """Start a service.

    Start a service.

    Usage: Basic CLI Use
      ```sh
      duploctl service start <service-name>
      ```
    
    Args:
      name (str): The name of the service to start.
      wait: Boolean flag to wait for service updates.

    Returns: 
      A success message if the service was started successfully.

    Raises:
      DuploError: If the service could not be started.
    """
    self.duplo.post(self.endpoint(f"ReplicationControllerstart/{name}"))
    if wait:
      service = self.find(name)
      self.wait(service, service)
    return {"message": f"Successfully started service '{name}'"}

  @Command()
  def pods(self, 
           name: args.NAME) -> dict:
    """Get the pods for a service.
    
    Args:
      name: The name of the service to get pods for.
    Returns: 
      message: A list of pods for the service.
    Raises:
      DuploError: If the service could not be found.
    """
    def controlled_by_service(pod):
      cb = pod.get("ControlledBy", None)
      same_name = pod["Name"] == name
      if cb is None:
        return same_name
      else:
        return same_name and cb.get("QualifiedType", None) == "kubernetes:apps/v1/ReplicaSet" 
    pods = self.__pod_svc.list()
    return [ pod for pod in pods if controlled_by_service(pod) ]
  
  @Command()
  def logs(self,
           name: args.NAME,
           wait: args.WAIT = False) -> dict:
    """Service Logs
    
    Get the logs for a service.

    Usage: Basic CLI Use
      ```sh
      duploctl service logs <service-name>
      ```
    Args:
      name (str): The name of the service to get logs for.
      wait (bool): Whether to wait for logs to update.
    Returns:
      message: A success message.
    """
    def show_logs():
      pods = self.pods(name)
      for pod in pods:
        self.__pod_svc.logs(pod=pod)
    if wait:
      try:
        while True:
          show_logs()
          time.sleep(3)
      except KeyboardInterrupt:
        pass
    else:
      show_logs()

  def current_replicaset(self, name: str):
    """Get the current replicaset for a service.
    
    Args:
      name (str): The name of the service to get replicaset for.
    Returns: 
      The current replicaset for the service.
    Raises:
      DuploError: If the service could not be found.
    """
    pods = self.pods(name)
    cb = pods[0].get("ControlledBy", None)
    if cb is None:
      return None
    return cb.get("NativeId", None)
  
  def image_from_body(self, body):
    """Get the image from a service body.
    
    Args:
      body (dict): The body of the service.
    Returns: 
      The image for the service.
    """
    tpl = body.get("Template", {})
    containers = tpl.get("Containers", [])
    for c in containers:
      if c["Name"] == body["Name"]:
        return c["Image"]
    else:
      return body.get("DockerImage", body.get("Image", None))

  def wait(self, old, updates):
    """Wait for a service to update."""
    name = old["Name"]
    cloud = old["Template"]["Cloud"]
    new_img = self.image_from_body(updates)
    old_img = self.image_from_body(old)
    new_replicas = updates.get("Replicas", None)
    replicas_changed = (old["Replicas"] != new_replicas) if new_replicas else False
    image_changed = (old_img != new_img) if new_img else False
    new_conf = updates.get("OtherDockerConfig", None)
    conf_changed = (old["Template"].get("OtherDockerConfig", None) != new_conf) if new_conf else False
    rollover = False
    if image_changed or conf_changed:
      rollover = True
    def check_pod_faults(pod, faults):
      """Check if the pod has any faults.
      
      This is for aws and gke because the faults are at the pod level.
      """
      for f in faults:
        if f["Resource"]["Name"] == pod["InstanceId"]:
          raise DuploFailedResource(f"Pod {pod['InstanceId']} raised a fault.\n{f['Description']}")
    def check_service_faults(faults):
      """Check if the service has any faults.
      
      This is only for azure because the faults are at the service level only
      """
      for f in faults:
        if f["ResourceName"] == name:
          raise DuploFailedResource(f"Service {name} raised a fault.\n{f['Description']}")
    def wait_check():
      svc = self.find(name)
      replicas = svc["Replicas"]
      # make sure the change has been applied
      if (image_changed and self.image_from_body(svc) != new_img):
        raise DuploError(f"Service {name} waiting for image update", 400)
      if (replicas_changed and replicas != new_replicas):
        raise DuploError(f"Service {name} waiting for replicas update", 400)
      if (conf_changed and svc["Template"].get("OtherDockerConfig", None) != new_conf):
        raise DuploError(f"Service {name} waiting for pod to update", 400)
      pods = self.pods(name)
      faults = self.tenant_svc.faults(id=self.tenant_id)
      running = 0

      # check for azure faults on service
      if cloud == 2:
        check_service_faults(faults)
      for p in pods:

        # azure doesn't have controlled by, so we skip this check if it's not there
        cb = p.get("ControlledBy", None) 
        # skip if the pod is not controlled by the new replicaset
        if (cloud != 2 and cb is not None) and (cb["NativeId"] == old.get("Replicaset", None) and rollover):
          continue

        # ignore this pod if the image is the old image
        img = p["Containers"][0]["Image"].removeprefix("docker.io/library/")
        if image_changed and img == old_img:
          continue

        # check for aws and gke faults on pod
        if cloud != 2:
          check_pod_faults(p, faults)

        # update total running pod count if one is running 
        if ((p["CurrentStatus"] == p["DesiredStatus"]) and p["DesiredStatus"] == 1):
          self.duplo.logger.warn(f"Pod {p['InstanceId']} is running")
          running += 1

      # make sure all the replicas are up
      if replicas != running:
        raise DuploError(f"Service {name} waiting for pods {running}/{replicas}", 400)
      
    # send to the base class to do the waiting
    super().wait(wait_check, 400, 11)

  def name_from_body(self, body):
    return body["Name"]

  @Command()
  def expose(self,
             name: args.NAME,
             container_port: args.CONTAINER_PORT=None,
             external_port: args.EXTERNAL_PORT=None,
             lb_type: args.LOAD_BALANCER_TYPE=None,
             protocol: args.PROTOCOL=None,
             visibility: args.LOAD_BALANCER_VISIBILITY="public",
             mode: args.LOAD_BALANCER_MODE="docker-mode",
             health_check_url: args.HEALTH_CHECK_URL=None) -> dict:
    """
    Expose a service.

    Usage: Basic CLI Use
      ```sh
      duploctl service expose <service-name> --lb-type applicationlb --container-port 80 --external-port 80  \
                             --visibility public --mode docker-mode --health-check-url / --protocol http
      ```

    Args:
        container-port (int): The internal port of the container to expose.
        external-port (int): The external port exposed by the load balancer. This is not used for targetgrouponly or k8clusterip load balancer types.
        lb-type (str): The load balancer type. Valid options are ['applicationlb', 'k8clusterip', 'k8nodeport', 'networklb', 'targetgrouponly'].
        protocol (str): The protocol to use, based on `lb_type`
          - applicationlb: http, https
          - networklb: tcp, udp, tls
          - targetgrouponly: http, https
          - k8clusterip: tcp, udp
          - k8nodeport: tcp, udp
        visibility (str): The load balancer visibility. Valid options are 'public' or 'private'.
        mode (str): The load balancer application mode. Valid options are 'docker-mode' or 'native-app'.
        health_check_url (str): The health check URL path. This must be empty for networklb, as it does not support health check paths.
    """
    service = self.find(name)
    tenant_id = service["TenantId"]
    lb_type_map = {
      "applicationlb": 1,
      "k8clusterip": 3,
      "k8nodeport": 4,
      "networklb": 6,
      "targetgrouponly": 7
    }

    if lb_type not in lb_type_map:
      raise DuploError(f"Invalid lb_type: {lb_type}. Must be one of {list(lb_type_map.keys())}")

    payload = {
        "LbType": lb_type_map[lb_type],
        "Port": container_port,
        "ExternalPort": external_port if external_port is not None else None,
        "IsInternal": visibility == "private",
        "IsNative": mode == "native-app",
        "Protocol": protocol,
        "HealthCheckUrl": health_check_url if lb_type != "networklb" else None,
        "ReplicationControllerName": name
    }

    self.duplo.post(f"subscriptions/{tenant_id}/LBConfigurationUpdate", payload)
    return {"message": f"Successfully exposed service '{name}'"}



from duplocloud.client import DuploClient
from duplocloud.errors import DuploError
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("rds")
class DuploRDS(DuploTenantResourceV3):
  """Resource for managing RDS instances."""
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, "aws/rds/instance")
    self.wait_timeout = 1200

  @Command()
  def create(self,
             body: args.BODY,
             wait: args.WAIT=False):
    """Create a DB instance.
    
    Args:
      body (dict): The body of the request.
    """
    name = self.name_from_body(body)
    s = None
    def wait_check():
      nonlocal s
      i = self.find(name)
      status = i.get("InstanceStatus", "submitted")
      if s != status:
        s = status
        self.duplo.logger.warn(f"DB instance {name} is {status}")
      if status != "available":
        raise DuploError(None)
    super().create(body, wait, wait_check)

  @Command()
  def find_cluster(self,
                   name: args.NAME):
    """Find a DB instance by name.
    
    Args:
      name (str): The name of the DB instance to find.
    Returns: 
      The DB instance object.
    Raises:
      DuploError: If the DB instance could not be found.
    """
    response = self.duplo.get(self.endpoint(name, "groupDetails"))
    return response.json()
  
  @Command()
  def stop(self,
           name: args.NAME,
           wait: args.WAIT=False):
    """Stop a DB instance."""
    def wait_check():
      i = self.find(name)
      if i["InstanceStatus"] in ["stopping","available"]:
        raise DuploError(f"DB instance {name} is still stopping")
    self.duplo.post(self.endpoint(name, "stop"))
    if wait:
      self.wait(wait_check, 1800, 10)
    return {
      "message": "DB instance stopped"
    }
  
  @Command()
  def start(self,
           name: args.NAME,
           wait: args.WAIT=False):
    """Start a DB instance."""
    def wait_check():
      i = self.find(name)
      if i["InstanceStatus"] in ["starting"]:
        raise DuploError(f"DB instance {name} is still starting")
    self.duplo.post(self.endpoint(name, "start"))
    if wait:
      self.wait(wait_check, 1800, 10)
    return {
      "message": "DB instance started"
    }
  
  @Command()
  def reboot(self,
             name: args.NAME,
             wait: args.WAIT=False):
    """Reboot a DB instance."""
    rebooting = False
    def wait_check():
      nonlocal rebooting
      i = self.find(name)
      if i["InstanceStatus"] == "rebooting" and not rebooting:
        rebooting = True
      if i["InstanceStatus"] == "available" and rebooting:
        return True # finally rebooting is a success
      raise DuploError(f"DB instance {name} is still rebooting")
    # Reboot the instance
    self.duplo.post(self.endpoint(name, "reboot"))
    if wait:
      self.wait(wait_check, 1800, 10)
    return {
      "message": "DB instance rebooted"
    }
  
  @Command()
  def set_instance_size(self,
                        name: args.NAME,
                        size: args.SIZE):
    """Set the size of a DB instance.
    
    Args:
      name (str): The name of the DB instance to update.
      size (str): The new size to use for the DB instance.
    """
    self.duplo.put(self.endpoint(name, "updatePayload"), {"SizeEx": size})
    return {
      "message": f"DB instance {name} resized to {size}"
    }
  
  @Command()
  def change_password(self,
                      name: args.NAME,
                      password: args.PASSWORD,
                      store: args.SAVE_SECRET):
    """Change the password of a DB instance.
    
    Args:
      name (str): The name of the DB instance to update.
      password (str): The new password to use for the DB instance.
    """
    body = {
      "Identifier": name,
      "MasterPassword": password,
      "StorePassword": store
    }
    self.duplo.post(self.endpoint(name, "changePassword"), body)
    return {
      "message": f"Password for DB instance {name} changed"
    }
  
  @Command()
  def set_monitor_interval(self,
                           name: args.NAME,
                           interval: args.INTERVAL,
                           immediate: args.IMMEDIATE=False):
    """Set the monitoring interval for a DB instance.
    
    Args:
      name (str): The name of the DB instance to update.
      interval (str): The new monitoring interval to use for the DB instance.
    """
    tenant_id = self.tenant["TenantId"]
    body = {
      "DBInstanceIdentifier": name,
      "ApplyImmediately": immediate, 
      "MonitoringInterval":interval
    }
    self.duplo.post(f"subscriptions/{tenant_id}/ModifyRDSDBInstance", body)
    return {
      "message": f"Monitoring interval for DB instance {name} set to {interval}"
    }
  
  @Command()
  def logging(self,
              name: args.NAME,
              enable: args.ENABLE):
    """Enable or disable logging for a DB instance."""
    self.duplo.put(self.endpoint(name, "updatePayload"), {"EnableLogging": enable})
    return {
      "message": f"DB instance {name} logging is {enable}"
    }
  
  @Command()
  def iam_auth(self,
               name: args.NAME,
               enable: args.ENABLE,
               immediate: args.IMMEDIATE=False):
    """Toggle IAM authentication for a DB instance."""
    body = {
      "DBInstanceIdentifier": name,
      "EnableIAMDatabaseAuthentication": enable,
      "ApplyImmediately": immediate
    }
    self.update(body)
    return {
      "message": f"IAM authentication for DB instance {name} is {enable}"
    }
  
  @Command()
  def final_snapshot(self,
                     name: args.NAME,
                     enable: args.ENABLE,
                     immediate: args.IMMEDIATE=False):
    """Toggle IAM authentication for a DB instance."""
    body = {
      "DBInstanceIdentifier": name,
      "SkipFinalSnapshot": not enable,
      "ApplyImmediately": immediate
    }
    self.update(body)
    return {
      "message": f"Final Snapshot for DB instance {name} is {enable}"
    }
  
  @Command()
  def snapshot(self,
               name: args.NAME):
    """Take a snapshot of a DB instance."""
    body = { "DBInstanceIdentifier": name }
    self.duplo.post(self.endpoint(name, "snapshot"), body)
    return {
      "message": f"Snapshot of DB instance {name} taken"
    }
  
  @Command()
  def restore(self,
              name: args.NAME,
              target: args.TARGET,
              time: args.TIME):
    """Restore a DB instance from a snapshot."""
    body = {
      "TargetName": target, 
      "RestoreTime": time
    }
    response = self.duplo.post(self.endpoint(name, "restorePointInTime"), body)
    return {
      "message": f"DB instance {name} restored to {target} at {time}",
      "data": response.json()
    }
  
  @Command()
  def retention_period(self,
                       name: args.NAME,
                       days: args.DAYS,
                       immediate: args.IMMEDIATE=False):
    """Set the retention period for a DB instance."""
    body = {
      "DBInstanceIdentifier": name,
      "BackupRetentionPeriod": days,
      "ApplyImmediately": immediate
    }
    self.update(body)
    return {
      "message": f"DB instance {name} retention period set to {days} days"
    }

  def name_from_body(self, body):
    other_name = body.get("DBInstanceIdentifier", None)
    name = body.get("Identifier", other_name)
    if not name.startswith("duplo"):
      name = "duplo" + name
    return name



import random
import pytest
import time

from duplocloud.errors import DuploError

class TestTenant:

  @pytest.mark.integration
  def test_listing_tenants(self, duplo):
    r = duplo.load("tenant")
    try:
      lot = r("list")
    except DuploError as e:
      pytest.fail(f"Failed to list tenants: {e}")
    # there is at least one tenant
    assert len(lot) > 0

  @pytest.mark.integration
  def test_finding_tenants(self, duplo):
    r = duplo.load("tenant")
    try:
      t = r("find", "default")
    except DuploError as e:
      pytest.fail(f"Failed to list tenants: {e}")
    assert t["AccountName"] == "default"

  @pytest.mark.integration
  @pytest.mark.dependency(name="create_tenant", scope='session')
  @pytest.mark.order(2)
  def test_creating_tenants(self, duplo, infra_name, e2e):
    t = duplo.load("tenant")
    name = duplo.tenant
    if not name:
      inc = random.randint(1, 100)
      name = f"duploctl{inc}"
      duplo.tenant = name
    # check if the tenant already exists
    try:
      print(f"Processing tenant '{name}'")
      i = t("find", name)
      print(f"Tenant '{name}' already exists")
      if i:
        pytest.skip(f"Tenant '{name}' already exists")
    except DuploError as e:
      pass
    try:
      t.create({
        "AccountName": name,
        "PlanID": infra_name,
        "TenantBlueprint": None
      }, wait=True)
      print(f"Tenant '{name}' created")
    except DuploError as e:
      pytest.fail(f"Failed to create tenant: {e}")
    time.sleep(180)

  @pytest.mark.integration
  @pytest.mark.dependency(name="delete_tenant", depends=["create_tenant"], scope='session')
  @pytest.mark.order(998)
  def test_find_delete_tenant(self, duplo):
    # now find it
    r = duplo.load("tenant")
    name = duplo.tenant
    print(f"Delete tenant '{name}'")
    try:
      nt = r("find", name)
      # now try to find again, but using the id this time
      nt = r.find(id=nt["TenantId"])
      assert nt["AccountName"] == name
    except DuploError as e:
      pytest.fail(f"Failed to find tenant {name}: {e}")
    # now delete the tenant
    try:
      r("config", name, "-D", "delete_protection")
      r("delete", name)
    except DuploError as e:
      pytest.fail(f"Failed to delete tenant: {e}")




import argparse
import logging
from .argtype import Arg, YamlAction, JsonPatchAction, DataMapAction
from .commander import available_resources, available_formats, VERSION

# the global args for the CLI

HOME_DIR = Arg('homedir', '--home-dir', 
            help='The home directory for duplo configurations',
            env='DUPLO_HOME')
"""Home Directory

Defaults to users home directory at `$HOME/.duplo`
This is where the cli will look by default for the config and cache as well. 
"""

CACHE_DIR = Arg('cachedir', '--cache-dir', 
            help='The cache directory for saved credentials.',
            env='DUPLO_CACHE')
"""Cache Directory

Defaults to `$HOME/.duplo/cache`. This is where the cli will store the cached credentials. Sometimes you may need to delete this directory to clear out old credentials. Simply type `duploctl` and this will print where the cache is currently stored.
"""

LOGLEVEL = Arg('log-level', '--loglevel', '-L',
            help='The log level to use.',
            default='INFO',
            env='DUPLO_LOG_LEVEL',
            choices=list(logging._nameToLevel.keys()))

CONFIG = Arg('configfile', '--config-file', 
            help='The path to the duploctl configuration file.',
            env='DUPLO_CONFIG')

CONTEXT = Arg("context", "--ctx",
              help='Use the specified context from the config file.',
              env='DUPLO_CONTEXT')

HOST = Arg('host', '-H', 
            help='The url to specified duplo portal.',
            env='DUPLO_HOST')

TOKEN = Arg('token', '-t', 
            help='The token to authenticate with duplocloud portal api.',
            env='DUPLO_TOKEN')

TENANT = Arg("tenant", "-T",
             help='The tenant name',
             env='DUPLO_TENANT')
"""Tenant Name

Scopes the command into the specified tenant. In the background the TENANT_ID is discovered using this name. So if TENANT_ID is set, this is ignored. Often times this is set as an environment variable so you don't have to choose the tenant each and every command. This can also be set in the config file within a context.
"""

TENANT_ID = Arg("tenantid", "--tenant-id", "--tid",
             help='The tenant id',
             env='DUPLO_TENANT_ID')
"""Tenant ID

Scopes the command into the specified tenant. This is the internal id of the tenant. If this is set, TENANT name argument is ignored.
"""

PLAN = Arg("plan", "-P",
            help='The plan name.',
            env='DUPLO_PLAN')
"""Plan Name

This is another high level placement style argument. This is used to scope the command to a specific plan aka infrastructure. 
"""

BODY = Arg("file", "-f", "--cli-input",
            help='A file to read the input from',
            type=argparse.FileType('r'),
            action=YamlAction)
"""File Body  

This is the file path to a file with the specified resources body within. Each Resource will have it's own schema for the body. This is a yaml/json file that will be parsed and used as the body of the request. View the docs for each individual resource to see the schema for the body.
"""

DATAMAP = Arg("fromfile","--from-file", "--from-literal",
            help='A file or literal value to add to the data map',
            action=DataMapAction)

DRYRUN = Arg("dryrun", "--dry-run",
            help='Do not submit any changes to the server',
            type=bool,
            action='store_true')

ARN = Arg("aws-arn", "--arn",
           help='The aws arn',
           default=None)

INTERACTIVE = Arg("interactive","-I", 
              help='Use interactive Login mode for temporary tokens. Do not use with --token.',
              type=bool,
              action='store_true')

ISADMIN = Arg("admin","--isadmin", 
              help='Request admin access when using interactive login.',
              type=bool,
              action='store_true')

NOCACHE = Arg("no-cache","--nocache", 
              help='Do not use cache credentials.',
              type=bool,
              action='store_true')

BROWSER = Arg("web-browser","--browser", 
              help='The desired web browser to use for interactive login',
              env='DUPLO_BROWSER',
              choices=[
                'chrome', 'chromium', 'firefox', 'safari', 'epiphany', 
                'edge', 'opera', 'konqueror', "kfm", 'w3m', 'lynx'
              ])
"""Web Browser

This is the browser of choice to use for interactive login. This is simply using the python [webbrowser](https://docs.python.org/3/library/webbrowser.html) module to open the browser. If you don't have a browser installed, you can use `w3m` or `lynx` for a text based browser.
"""

OUTPUT = Arg("output", "-o",
              help='The output format',
              default='json',
              env='DUPLO_OUTPUT',
              choices=available_formats())

QUERY = Arg("query", "-q",
            help='The jmespath query to run on a result')

PATCHES = Arg("patches", "--add", "--remove", "--copy", "--replace", "--test", "--move",
              help='The json patch to apply',
              action=JsonPatchAction)

VERSION = Arg("version", "--version",
              action='version', 
              version=f"%(prog)s {VERSION}",
              type=bool)

EXCLUDE = Arg("exclude", '--exclude',
              action='append',
              help='Exclude from the command')

# The rest are resource level args for commands
SERVICE = Arg('service', 
              help='The service to run',
              choices=available_resources())

COMMAND = Arg('command', 
             help='The subcommand to run')

# generic first positional arg for resource name
NAME = Arg("name", 
            nargs='?',
            help='The resource name')

IMAGE = Arg("image", 
            help='The image to use')

S3BUCKET = Arg("bucket",
            help='The s3 bucket to use')

S3KEY = Arg("key",
            help='The s3 key to use')

HEALTH_CHECK_URL = Arg("health-check-url", "--health-check-url",
            help='The health check URL')

CONTAINER_PORT = Arg("container-port", "--container-port",
            help='Container port')

EXTERNAL_PORT = Arg("external-port", "--external-port",
            help='The external port')

PROTOCOL = Arg("protocol", "--protocol",
            help='The protocol',
            choices=['http','https','tcp','udp','tls'])

LOAD_BALANCER_TYPE = Arg("lb-type", "--lb-type",
            help='The type of load balancer. Valid options are [applicationlb, k8clusterip, k8nodeport, networklb, targetgrouponly].',
            choices=['applicationlb', 'k8clusterip', 'k8nodeport', 'networklb', 'targetgrouponly'])

LOAD_BALANCER_VISIBILITY = Arg("visibility", "--visibility",
            help='The visibility of load balancer. Valid options are \"public\" or \"private\".  Default is public.',
            choices=['public', 'private'],
            default = 'public')

LOAD_BALANCER_MODE = Arg("mode", "--mode",
            help='The mode of load balancer. Valid options are \"docker-mode\" or \"native-app\".  Default is docker-mode.',
            choices=['docker-mode', 'native-app'],
            default = 'docker-mode')

SERVICEIMAGE = Arg("serviceimage", "-S",
            help='takes two arguments, a service name and an image:tag',
            action='append',
            nargs=2,
            metavar=('service', 'image'))

SETVAR = Arg("setvar", "-V",
            help='a key and value to set as an environment variable',
            action='append',
            nargs=2,
            metavar=('key', 'value'))

STRATEGY = Arg("strategy", "-strat",
            help='The merge strategy to use. Valid options are \"merge\" or \"replace\".  Default is merge.',
            choices=['merge', 'replace'],
            default = 'merge')

DELETEVAR = Arg("deletevar", "-D",
            action='append',
            help='a key to delete from the environment variables')

SCHEDULE = Arg("schedule","-s", 
               help='The schedule to use')

CRONSCHEDULE = Arg("cronschedule", 
               help='The schedule to use')

ENABLE = Arg("enable","-y", 
              help='Enable or disable the feature',
              type=bool,
              action=argparse.BooleanOptionalAction)

MIN = Arg("min", "-m",
          help='The minimum number of replicas',
          type=int)

MAX = Arg("max", "-M",
          help='The maximum number of replicas',
          type=int)

REPLICAS = Arg("replicas", "-r",
               help = 'Number of replicas for service',
               type = int)

WAIT = Arg("wait", "-w",
           help='Wait for the operation to complete',
           type=bool,
           action='store_true')

SIZE = Arg("size",
           help='The instance size to use')

SAVE_SECRET = Arg("save-secret", "--save",
                 help='Save the secret to secrets manager.',
                 type=bool,
                 action='store_true')

PASSWORD = Arg("password",
                help='The password to use')

INTERVAL = Arg("interval",
                help='The monitoring interval to use',
                type=int,
                choices=[1, 5, 10, 15, 30, 60])

IMMEDIATE = Arg("immediate", "-i",
                help='Apply the change immediately',
                type=bool,
                action='store_true')

TARGET = Arg("target", "--target-name",
             help='The target name to use')

TIME = Arg("time", "--time",
           help='The time to use')

DAYS = Arg("days", 
            help='The days to use',
            type=int)

CONTENT_DIR = Arg('content', '--content-dir', 
            help='The content directory for a website.',
            default='dist',
            env='DUPLO_CONTENT')

SSM_PARAM_TYPE = Arg("parametertype", "-ptype",
            help='The type of parameter to create',
            choices=['String', 'StringList', 'SecureString'],
            default='String')

PARAM_CONTENT = Arg('parametervalue', '-pval',
            help='Arbitrary text to add to the parameter')

SHOW_SENSITIVE = Arg('showsensitive', '-show',
            help='Return sensitive values to output. WARNING - ENABLING THIS SETTING MAY DISPLAY SENSITIVE DATA TO STDOUT/LOG FILES',
            type=bool,
            action='store_true')
            


from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV2
from duplocloud.errors import DuploError
from duplocloud.commander import Command, Resource
import duplocloud.args as args


@Resource("ecs")
class DuploEcsService(DuploTenantResourceV2):

  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)

  @Command()
  def list(self):
    """Retrieve a list of all ECS services in a tenant."""
    tenant_id = self.tenant["TenantId"]
    url = f"subscriptions/{tenant_id}/GetEcsServices"
    response = self.duplo.get(url)
    return response.json()
  
  @Command()
  def find(self, 
           name: args.NAME):
    """Find a ECS service by name.

    Args:
      name (str): The name of the ECS service to find.
    Returns:
      The ECS service object.
    Raises:
      DuploError: If the ECS service could not be found.
    """
    try:
      return [s for s in self.list() if s["Name"] == name][0]
    except IndexError:
      raise DuploError(f"ECS Service '{name}' not found", 404)
  
  @Command()
  def list_task_def_family(self) -> dict:
    """List ECS Task Definitions
    
    Retrieve a list of all ECS task definitions in a tenant.

    Example: 
      CLI usage  
      ```sh  
      duploctl ecs list_definitions
      ```

    Returns:
      task_def_family: A list of ECS task definitions.
    """
    tenant_id = self.tenant["TenantId"]
    path = f"v3/subscriptions/{tenant_id}/aws/ecs/taskDefFamily"
    response = self.duplo.get(path)
    return response.json()
    
  @Command()
  def find_def(self, 
               name: args.NAME):
    """Find a ECS task definition by name.

    Args:
      name (str): The name of the ECS task definition to find.
    Returns:
      The ECS task definition object.
    Raises:
      DuploError: If the ECS task definition could not be found.
    """
    tdf = self.find_task_def_family(name)
    arn = tdf["VersionArns"][-1]
    return self.find_def_by_arn(arn)
  
  @Command()
  def find_def_by_arn(self, 
                      arn: args.ARN):
    """Find a ECS task definition by ARN.

    Args:
      arn (str): The ARN of the ECS task definition to find.
    Returns:
      The ECS task definition object.
    Raises:
      DuploError: If the ECS task definition could not be found.
    """
    tenant_id = self.tenant["TenantId"]
    path = f"subscriptions/{tenant_id}/FindEcsTaskDefinition"
    response = self.duplo.post(path, {"Arn": arn})
    return response.json()
      
  @Command()
  def find_task_def_family(self, 
                           name: args.NAME):
    """Find a ECS task definition by name.

    Args:
      name (str): The name of the ECS task definition to find.
    Returns:
      The ECS task definition object.
    Raises:
      DuploError: If the ECS task definition could not be found.
    """
    tenant_id = self.tenant["TenantId"]
    path = f"v3/subscriptions/{tenant_id}/aws/ecs/taskDefFamily/{name}"
    response = self.duplo.get(path)
    return response.json()
  
  @Command()
  def update(self, 
             body: args.BODY):
    """Update an ECS service.

    Args:
      body (dict): The updated ECS service object.
    Returns:
      The updated ECS object.
    Raises:
      DuploError: If the ECS service could not be updated.
    """
    tenant_id = self.tenant["TenantId"]
    path = f"subscriptions/{tenant_id}/UpdateEcsService"
    self.duplo.post(path, body)
    return {"message": "ECS Service updated"}
  
  @Command()
  def update_taskdef(self, 
                     body: args.BODY):
    """Update an ECS task definition.

    Args:
      body (dict): The updated ECS task definition object.
    Returns:
      The updated ECS object.
    Raises:
      DuploError: If the ECS task definition could not be updated.
    """
    tenant_id = self.tenant["TenantId"]
    path = f"subscriptions/{tenant_id}/UpdateEcsTaskDefinition"
    b = self.__ecs_task_def_body(body)
    response = self.duplo.post(path, b)
    return {"arn": response.json()}

  @Command()
  def update_image(self, 
                   name: args.NAME, 
                   image: args.IMAGE) -> dict:
    """Update the image for an ECS service.

    Example:
      CLI usage
      ```sh
      duploctl ecs update_image my-service my-image
      ```

    Args:
      name: The name of the ECS service to update.
      image: The new image to use.
    Returns:
      ecs: The updated ECS object.
    Raises:
        DuploError: If the ECS service could not be updated.
    """
    svc = self.find(name)
    tdf = self.find_def_by_arn(svc["TaskDefinition"])
    tdf["ContainerDefinitions"][0]["Image"] = image
    arn = self.update_taskdef(tdf)["arn"]
    svc["TaskDefinition"] = arn
    return self.update(svc)

  def __ecs_task_def_body(self, task_def):
    containers = [
      self.__ecs_container_update_body(c) 
      for c in task_def.get("ContainerDefinitions", [])
    ]
    return {
      "Family": task_def["Family"],
      "Cpu": task_def["Cpu"],
      "Memory": task_def["Memory"],
      "InferenceAccelerators": task_def.get("InferenceAccelerators", []),
      "NetworkMode": task_def.get("NetworkMode", {}),
      "ContainerDefinitions": containers
    }
  
  def __ecs_container_update_body(self, container_def):
    update_body = {
        "Essential": container_def.get("Essential"),
        "Image": container_def.get("Image") ,
        "Name": container_def.get("Name") ,
        "PortMappings": container_def.get("PortMappings", []) ,
        "Environment": container_def.get("Environment", {}) ,
        "Command": container_def.get("Command", {}) ,
        "Secrets": container_def.get("Secrets", {}) ,
    }
    
    # Add LogConfiguration only if it exists in container_def
    if "LogConfiguration" in container_def:
        update_body["LogConfiguration"] = container_def["LogConfiguration"]
    # Add FirelensConfiguration only if it exists in container_def
    if "FirelensConfiguration" in container_def:
        update_body["FirelensConfiguration"] = container_def["FirelensConfiguration"]
    return update_body



from duplocloud.client import DuploClient
from duplocloud.errors import DuploError
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("configmap")
class DuploConfigMap(DuploTenantResourceV3):
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, "k8s/configmap")

  @Command()
  def create(self, 
             name: args.NAME=None,
             body: args.BODY=None,
             data: args.DATAMAP=None,
             dryrun: args.DRYRUN=False,
             wait: args.WAIT=False) -> dict:
    """Create a Configmap resource.

    Usage: CLI Usage
      ```sh
      duploctl configmap create -f 'configmap.yaml'
      ```
      Contents of the `configmap.yaml` file
      ```yaml
      --8<-- "src/tests/data/configmap.yaml"
      ```

    Example: One liner example
      ```sh
      echo \"\"\"
      --8<-- "src/tests/data/configmap.yaml"
      \"\"\" | duploctl configmap create -f -
      ```
    
    Args:
      name: The name to set the configmap to.
      body: The resource to create.
      data: The data to add to the configmap.
      dryrun: Do not submit any changes to the server.
      wait: Wait for the resource to be created.

    Returns: 
      message: Success message.

    Raises:
      DuploError: If the resource could not be created.
    """
    if not name and not body:
      raise DuploError("Name is required when body is not provided")
    if not body:
      body = {}
    # make sure the body has a metadata key
    if 'metadata' not in body:
      body['metadata'] = {}
    # also make sure the data key is present
    if 'data' not in body:
      body['data'] = {}
    if name:
      body['metadata']['name'] = name
    if data:
      body['data'].update(data)
    if dryrun:
      return body
    else:
      return super().create(body, wait=wait)





class DuploError(Exception):
  """Base class for Duplo errors."""
  def __init__(self, message, code=1, response=None):
    super().__init__(message)
    self.message = message
    self.code = code
    self.response = response
  def __str__(self):
    return self.message

class DuploExpiredCache(DuploError):
  """Raised when the Duplo cache is expired."""
  def __init__(self, key: str):
    self.key = key
    super().__init__("Cache item {key} is expired", 404)

class DuploFailedResource(DuploError):
  """Raised when a Duplo resource is in a failed state."""
  def __init__(self, name: str):
    super().__init__(f"{name} is in a failed state", 412)



from datetime import timedelta
import datetime
from duplocloud.client import DuploClient
from duplocloud.resource import DuploResource
from duplocloud.errors import DuploError
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("tenant")
class DuploTenant(DuploResource):
  """Duplo Tenant Resource
  
  The tenant resource provides a set of commands to manage tenants in the Duplo system.

  Usage: Basic CLI Use  
    ```sh
    duploctl tenant <action>
    ```
  """
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
  
  @Command("ls")
  def list(self):
    """List Tenants
    
    Retrieve a list of all tenants in the Duplo system. 

    Usage: Basic CLI Use
      ```bash
      duploctl tenant list
      ```
    
    Returns:
      tenants (list): A list of tenants.
    """
    response = self.duplo.get("adminproxy/GetTenantNames")
    return response.json()
  
  @Command()
  def list_users(self, 
               name: args.NAME) -> dict:
    """List users assigned to a tenant
    
    Retrieve a list of all users with access to a tenant

    Usage: Basic CLI Use
      ```bash
      duploctl tenant list_users
      ```
    
    Returns:
      users (list): A list of users with access to the tenants, their readonly status, and if they're an admin user
    """
    tenant = self.find(name)
    tenant_id = tenant["TenantId"]
    response = self.duplo.get("admin/GetAllTenantAuthInfo")
    tenant_users = []
    for tenant in response.json():
        if tenant["TenantId"] == tenant_id:
            for user in tenant['UserAccess']:
                tenant_users.append({
                    "Username": user['Username'],
                    "IsReadOnly": f"{user['Policy']['IsReadOnly']}",
                    "IsAdmin": "False"
                })

    # Admins have access to all tenants. Check for them and add them
    users = self.duplo.get("admin/GetAllUserRoles")
    for user in users.json():
        if "Administrator" in user['Roles']:
            # If the user is already in the list for the tenant, mark them as admins. This shouldn't be possible.
            existing_user = next((u for u in tenant_users if u['Username'] == user['Username']), None)
            if existing_user:
              for user in tenant_users:
                  if user == existing_user:
                      user["IsAdmin"] = True
                      break
            else:
                tenant_users.append({
                    "Username": user['Username'],
                    "IsReadOnly": "False",
                    "IsAdmin": "True"
                })
    
    return tenant_users

  @Command("get")
  def find(self, 
           name: args.NAME=None,
           id: str=None) -> dict:
    """Find a tenant.

    Find a tenant by name or id. Passing in a name directly takes highest precedence.
    If a name is not passed in, the id is second highest precedence. Lastly if the global
    tenant name is set, that will be used. 

    The global tenant id takes care of the commandline. For other code, sometimes the id 
    needs to be passed in directly. If this happens, that id takes most precedence.

    Usage: Basic CLI Use
      ```bash
      duploctl tenant find <name>
      ```

    Args:
      name: The name or id of the tenant to find.
      id: The id of the tenant to find. Optional and code only.
    
    Returns:
      tenant: The tenant.
    """
    key = None 
    ref = None 
    if id or (not name and self.duplo.tenantid):
      key = "TenantId"
      ref = id or self.duplo.tenantid
    else:
      key = "AccountName"
      ref = name.lower() if name else self.duplo.tenant
    try:
      return [t for t in self.list() if t[key] == ref][0]
    except IndexError:
      raise DuploError(f"Tenant '{ref}' not found", 404)
  
  @Command()
  def create(self, 
             body: args.BODY,
             wait: args.WAIT=False) -> dict:
    """Create Tenant.
    
    Create a new tenant with a new body for a tenant. 

    Usage: Basic CLI Use
      ```bash
      duploctl tenant create --file tenant.yaml
      ```

    Example: Tenant Body
      Contents of the `tenant.yaml` file
      ```yaml
      --8<-- "src/tests/data/tenant.yaml"
      ```

    Example: Create One Liner
      Here is how to create a tenant in one line.  
      ```bash
      echo \"\"\"
      --8<-- "src/tests/data/tenant.yaml"
      \"\"\" | duploctl tenant create -f -
      ```
    
    Args:
      body: The body of the tenant to create.
      wait: Wait for the tenant to be created.

    Returns:
      message: The message that the tenant was created
    """
    name = body["AccountName"]
    self.duplo.post("admin/AddTenant", body)
    def wait_check():
      self.find(name)
    if wait:
      self.wait(wait_check)
    return {
      "message": f"Tenant '{name}' created"
    }
  
  @Command()
  def delete(self,
             name: args.NAME=None) -> dict:
    """Delete Tenant

    Delete a tenant by name.

    Usage: Basic CLI Use
      ```sh
      duploctl tenant delete <name>
      ```
    
    Args:
      name: The name of the tenant to delete.
    
    Returns:
      message: The message that the tenant was deleted.
    """
    tenant = self.find(name)
    tenant_id = tenant["TenantId"]
    self.duplo.post(f"admin/DeleteTenant/{tenant_id}", None)
    return {
      "message": f"Tenant '{name}' deleted"
    }
    
  @Command()
  def shutdown(self, 
               name: args.NAME=None, 
               schedule: args.SCHEDULE=None) -> dict:
    """Shutdown Tenant
    
    Shutdown a tenant by name and with a schedule.

    Usage: Basic CLI Use
      ```bash
      // Below command shutdown the tenant after 5 minutes (default)
      duploctl tenant shutdown <tenant-name>
      // Below command shutdown the tenant after given time 'minutes'(m), 'hours'(h) and 'day'(d) and it also support overriding the shutdown time.
      duploctl tenant shutdown <tenant-name> <time) // Example: 5m, 2h, 1d
      ```

    Args:
      name: The name of the tenant to shutdown.
      schedule: The schedule to shutdown the tenant.
    
    Returns:
      message: The message that the tenant was shutdown
    """
    tenant = self.find(name)
    tenant_id = tenant["TenantId"]
    # if the schedule not specified then set the date 5 minute from now
    if schedule is None:
      now = datetime.datetime.now() + datetime.timedelta(minutes=5)
      schedule = now.strftime("%a, %d %b %Y %H:%M:%S GMT")
    else:
      interval, unit = int(schedule[:-1]), schedule[-1]
      # Calculate the timedelta based on the specified time interval
      if unit == 'm':
        delta = timedelta(minutes=interval)
      elif unit == 'h':
        delta = timedelta(hours=interval)
      elif unit == 'd':
        delta = timedelta(days=interval)
      else:
          raise ValueError("Invalid time unit specified. Please use 'm' for minutes, 'h' for hours, or 'd' for days.")

      current_time = datetime.datetime.utcnow()
      # Calculate the future time after adding the timedelta
      future_time = current_time + delta
      # Format the future time in the desired string format
      schedule = future_time.strftime("%a, %d %b %Y %H:%M:%S GMT")

    res = self.duplo.post("adminproxy/UpdateTenantCleanupTimers", {
      "TenantId": tenant_id,
      "PauseTime": schedule
    })

    if res.status_code == 200:
      return {
        "message": f"Tenant '{name}' will shutdown on {schedule}"
      }
    else:
      raise DuploError(f"Failed to expire tenant '{name}'", res.status_code)

  @Command()
  def logging(self, 
              name: args.NAME=None, 
              enable: args.ENABLE=True) -> dict:
    """Toggle Loggine
    
    Enable or disable logging for a tenant.

    Usage: Basic CLI Use
      ```bash
      duploctl tenant logging <tenant-name> (default: true) // false not supported
      ```
    
    Args:
      name: The name of the tenant to toggle logging.
      enable: Enable or disable logging.
    
    Returns:
      message: The message that the tenant logging was toggled
    """
    tenant = self.find(name)
    tenant_id = tenant["TenantId"]
    # add or update the tenant in the list of enabled tenants
    response = self.duplo.get("admin/GetLoggingEnabledTenants")
    log_tenants = response.json()
    for t in log_tenants:
      if t["TenantId"] == tenant_id:
        t["Enabled"] = enable
        break
    else:
      log_tenants.append({"TenantId": tenant_id, "Enabled": enable})
    print(log_tenants)
    # update the entire list
    res = self.duplo.post("admin/UpdateLoggingEnabledTenants", log_tenants)
    if res.status_code == 200:
      return {
        "message": f"Tenant '{name}' logging {enable}"
      }
    else:
      raise DuploError(f"Failed to {'enable' if enable else 'disable'} tenant '{name}'", res.status_code)

  @Command()
  def billing(self,
              name: args.NAME=None) -> dict:
    """Tenant Billing Information
    
    Get the spend for the tenant.

    Usage: Basic CLI Use
      ```bash
      duploctl tenant billing <tenant-name>
      ```

    Args:
      name: The name of the tenant to get billing information for.

    Returns:
      billing: The billing information for the tenant.
    """
    tenant = self.find(name)
    tenant_id = tenant["TenantId"]
    response = self.duplo.get(f"v3/billing/subscriptions/{tenant_id}/aws/billing")
    return response.json()
  
  @Command()
  def config(self,
             name: args.NAME=None,
             setvar: args.SETVAR=[],
             deletevar: args.DELETEVAR=[]) -> dict:
    """Manage Tenant Settings
    
    Send a series of new settings and even some to delete.

    Usage: Basic CLI Use
      ```bash
      duploctl tenant config <tenant-name> --setvar <key> <value> --deletevar key3
      ```

    Args:
      name: The name of the tenant to manage.
      setvar: A series of key value pairs to set.
      deletevar: The keys to delete.
    
    Returns:
      message: The message that the tenant settings were updated.
    """
    updates = []
    creates = []
    changes = []
    tenant = self.find(name)
    tenant_id = tenant["TenantId"]
    endpoint = f"v3/admin/tenant/{tenant_id}/metadata"
    curr_settings = tenant.get("Metadata", [])
    curr_keys = [s["Key"] for s in curr_settings]
    # flatten k/v pair while dedupe and remove deleted keys
    new_settings = {s[0]: s[1] for s in setvar if s[0] not in deletevar}
    # only update if the value is different and create if the key is not present
    for k, v in new_settings.items():
      s = {"Key": k, "Value": v}
      for c in curr_settings:
        if c["Key"] == k:
          if c["Value"] != v:
            updates.append(s)
          break
      else:
        creates.append(s)
    # create, update, and delete the settings
    for s in creates:
      response = self.duplo.post(endpoint, s)
      change = response.json()
      change["Operation"] = "create"
      changes.append(change)
    for s in updates:
      response = self.duplo.put(f"{endpoint}/{s['Key']}", s)
      change = response.json()
      change["Operation"] = "update"
      changes.append(change)
    for k in deletevar:
      if k in curr_keys:
        self.duplo.delete(f"{endpoint}/{k}")
        change = {"Key": k, "Operation": "delete"}
        changes.append(change)
    return {
      "message": f"Successfully updated settings for tenant '{name}'",
      "changes": changes
    }
    
  @Command()
  def host_images(self,
                  name: args.NAME = None) -> list:
    """Available Duplo Host Images
    
    Get the list of host images for the tenant. These AMI's are region scoped. 

    Usage: Basic CLI Use
      ```bash
      duploctl tenant host_images <tenant-name>
      ```

    Args:
      name: The name of the tenant to get host images for.

    Returns:
      host_images: A list of host images.
    """
    tenant = self.find(name)
    tenant_id = tenant["TenantId"]
    response = self.duplo.get(f"v3/subscriptions/{tenant_id}/nativeHostImages")
    return response.json()

  @Command()
  def faults(self,
             name: args.NAME = None,
             id: str = None) -> list:
    """Tenant Faults
    
    Retrieves the list of faults for a tenant.

    Usage: Basic CLI Use
      ```bash
      duploctl tenant faults <tenant-name>
      ```

    Args:
      name: The name of the tenant to get faults for.
      id: The id of the tenant to get faults for. Optional and code only.
    
    Returns:
      faults: A list of faults.
    """
    tenant = self.find(name, id)
    tenant_id = tenant["TenantId"]
    response = self.duplo.get(f"subscriptions/{tenant_id}/GetFaultsByTenant")
    return response.json()
  
  @Command()
  def region(self,
             name: args.NAME = None) -> dict:
    """Tenant Region
    
    Get the region the tenants infrastructure is placed in.

    Usage: Basic CLI Use
      ```bash
      duploctl tenant region <tenant-name>
      ```
    
    Args:
      name: The name of the tenant to get the region for.

    Returns:
      region: The region the tenant is in.
    """
    tenant = self.find(name)
    tenant_id = tenant["TenantId"]
    response = self.duplo.get(f"subscriptions/{tenant_id}/GetAwsRegionId")
    return {
      "region": response.json()
    }

  @Command()
  def start(self, 
            name: args.NAME = None, 
            wait: args.WAIT=False, 
            exclude: args.EXCLUDE=None) -> dict:
    """Start Tenant All Resources

    Starts all resources of a tenant.

    Usage: Basic CLI Use
      ```bash
      duploctl tenant start
      ```

    Args:
      wait: Wait for the resources to start.
      exclude (optional): A list of resources to exclude from starting. Can include:
        - hosts/<host_name>: Exclude a specific host.
        - rds/<rds_name>: Exclude a specific RDS instance.
        - hosts/at/<allocation_tags>: Exclude hosts with specific allocation tags.

    Returns:
      message: A success message.
    """
    service_types = {"hosts": [], "rds": []}
    host_at = []
    if exclude:
      for item in exclude:
        category, value = item.split('/', 1)
        if 'at/' in value:
          _, at_name = value.split('at/', 1)
          host_at.append(at_name)
        elif category in {"hosts", "rds"}:
          tenant_name = self.find(name)['AccountName']
          if category == "hosts":
            prefix = f"duploservices-{tenant_name}"
            value = f"{prefix}-{value}" if not value.startswith(prefix) else value
          elif category == "rds":
            value = f"duplo{value}" if not value.startswith("duplo") else value
          service_types[category].append(value)
        else:
          print(f"Unknown service: {category}")

    host_at_exclude = self.get_hosts_to_exclude(host_at)
    service_types['hosts'] = list(set(service_types['hosts']) | set(host_at_exclude))

    for service_type in service_types.keys():
      service = self.duplo.load(service_type)
      for item in service.list():
        service_name = service.name_from_body(item)
        if service_name not in service_types[service_type]:
          service.start(service_name, wait)
    return {
      "message": "Successfully started all resources for tenant"
    }

  @Command()
  def stop(self, 
           name: args.NAME = None, 
           wait: args.WAIT=False, 
           exclude: args.EXCLUDE=None) -> dict:
    """Stop Tenant All Resources

    Stops all resources of a tenant.

    Usage: Basic CLI Use
      ```bash
      duploctl tenant stop
      ```

    Args:
      wait: Wait for the resources to stop.
      exclude (optional): A list of resources to exclude from stopping. Can include:
        - hosts/<host_name>: Exclude a specific host.
        - rds/<rds_name>: Exclude a specific RDS instance.
        - hosts/at/<allocation_tags>: Exclude hosts with specific allocation tags.

    Returns:
      message: A success message.
    """
    service_types = {"hosts": [], "rds": []}
    host_at = []
    if exclude:
      for item in exclude:
        category, value = item.split('/', 1)
        if 'at/' in value:
          _, at_name = value.split('at/', 1)
          host_at.append(at_name)
        elif category in {"hosts", "rds"}:
          tenant_name = self.find(name)['AccountName']
          if category == "hosts":
            prefix = f"duploservices-{tenant_name}"
            value = f"{prefix}-{value}" if not value.startswith(prefix) else value
          elif category == "rds":
            value = f"duplo{value}" if not value.startswith("duplo") else value
          service_types[category].append(value)
        else:
          print(f"Unknown service: {category}")

    host_at_exclude = self.get_hosts_to_exclude(host_at)
    service_types['hosts'] = list(set(service_types['hosts']) | set(host_at_exclude))

    for service_type in service_types.keys():
      service = self.duplo.load(service_type)
      for item in service.list():
        service_name = service.name_from_body(item)
        if service_name not in service_types[service_type]:
          service.stop(service_name, wait)
    return {
      "message": "Successfully stopped all resources for tenant"
    }

  def get_hosts_to_exclude(self, host_at):
    host_at_exclude = []
    hosts = self.duplo.load('hosts').list()
    for host in hosts:
      allocation_tags_value = ''
      minion_tags = host.get('MinionTags', [])
      for tag in minion_tags:
        if tag.get('Key') == 'AllocationTags':
          allocation_tags_value = tag.get('Value')

      if allocation_tags_value in host_at:
        host_at_exclude.append(host['FriendlyName'])
    return host_at_exclude

  @Command()
  def dns_config(self, 
                 name: args.NAME=None) -> dict:
    """Tenant DNS Config

    Retrieve DNS configuration for a tenant by name..

    Usage: Basic CLI Use
      ```sh
      duploctl tenant dns_config <name>
      ```

    Args:
      name: The name of the tenant.

    Returns:
      dns_config: A dictionary containing the DNS configuration of the tenant.
    """
    tenant = self.find(name)
    tenant_id = tenant["TenantId"]
    response = self.duplo.get(f"v3/subscriptions/{tenant_id}/aws/dnsConfig")
    return response.json()

  @Command()
  def add_user(self, 
               name: args.NAME) -> dict:
    """Add User to Tenant
    
    Usage: CLI Usage
      ```sh
      duploctl tenant add_user <user> --tenant <tenant_name>
      ```

    Args:
      name: The name of the user to add to the tenant.

    Returns:
      message: A message indicating the user was added to the tenant.
    """
    tenant_id = self.find(self.duplo.tenant)["TenantId"]
    res = self.duplo.post("admin/UpdateUserAccess", {
      "Policy": { "IsReadOnly": None },
      "Username": name,
      "TenantId": tenant_id
    })
    # check http response is 204
    if res.status_code != 204:
      raise DuploError(f"Failed to add user '{name}' to tenant '{self.duplo.tenant}'", res["status_code"])
    else:
      return f"User '{name}' added to tenant '{self.duplo.tenant}'"
    
  @Command()
  def remove_user(self, 
                 name: args.NAME) -> dict:
    """Remove a User from a Tenant
    
    Usage: CLI Usage
      ```sh
      duploctl tenant remove_user <user> --tenant <tenant_name>
      ```

    Args:
      name: The name of the user to remove from the tenant.

    Returns:
      message: A message indicating the user was removed from the tenant.
    """
    tenant_id = self.find(self.duplo.tenant)["TenantId"]
    res = self.duplo.post("admin/UpdateUserAccess", {
      "Policy": {},
      "Username": name,
      "TenantId": tenant_id,
      "State": "deleted"
    })

    # check http response is 204
    if res.status_code != 204:
      raise DuploError(f"Failed to remove user '{name}' from tenant '{self.duplo.tenant}'", res["status_code"])
    else:
      return f"User '{name}' removed from tenant '{self.duplo.tenant}'"
    


from duplocloud import args
from duplocloud.client import DuploClient
from duplocloud.errors import DuploError
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Command, Resource

@Resource("ssm_param")
class DuploParam(DuploTenantResourceV3):
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, "aws/ssmParameter")
  
  def name_from_body(self, body):
    return body["Name"]
  
  @Command()
  def create(self, 
             name: args.NAME=None,
             body: args.BODY=None,
             paramtype: args.SSM_PARAM_TYPE=None,
             value: args.PARAM_CONTENT=None,
             dryrun: args.DRYRUN=False,
             wait: args.WAIT=False) -> dict:
    """Create an SSM Parameter
    Usage: cli usage
      ```sh
      duploctl ssm_param create <name> -pval <value> -ptype <String|SecureString|StringList>
      ```
    
    Args:
      name: The name of the SSM Parameter to create.
      -ptype/--parametertype: The type of parameter to create, must be String, SecureString, or StringList
      -pval/--parametervalue: Arbitrary text to set in the parameter.  StringList expects comma separated values.
      -body: path to a raw json/yaml post body, e.g:
      ```
      {
        "Type": "String",
        "Value": "myvalue",
        "Name": "MyStringParameter"
      }
      ```

    Returns: 
      resource: The SSM Parameter object.
      
    Raises:
      DuploError: If the SSM Parameter already exists.
    """
    # Make sure the user passes name and value, or a body (from file). Paramtype is defaulted to string.
    if not body and (not name or not value):
      raise DuploError("name and parameter value are required when body is not provided")
    if not body:
      body = {}
    # If the user passed in settings, use them.  Otherwise use what's in the file.
    if paramtype:
      body['Type'] = paramtype
    if name:
      body['Name'] = name
    if value:
      body['Value'] = value
    if dryrun:
      return body
    else:
      return super().create(body, wait=wait)

  @Command()
  #Implement find with opt-in option to display sensitive data for secureString params. 
  #Note that using find on SecureString params still returns the sensitive data!
  #This is only to protect against accidental sensitive data exposure.
  def find(self, 
           name: args.NAME,
           show_sensitive: args.SHOW_SENSITIVE=False) -> dict:
    """Find SSM Parameter resources by name and return it's content

    Usage: cli usage
      ```sh
      duploctl ssm_param find <name>
      ```
    
    Args:
      name: The name of the SSM Parameter to find.
      -show/--showsensitive: Display value of SecureString parameters

    Returns: 
      resource: The SSM Parameter object.
      
    Raises:
      DuploError: If the SSM Parameter could not be found.
    """
    response = self.duplo.get(self.endpoint(name))
    if response.json()['Type']=="SecureString" and not show_sensitive:
      obfuscated_response=response.json()
      sensitive_len=len(response.json()["Value"])
      placeholder="*"
      obfuscated_response["Value"]=placeholder * sensitive_len
      return obfuscated_response
    else:
      return response.json()
    
  @Command()
  def update(self, 
             name: args.NAME=None,
             strategy: args.STRATEGY='merge',
             value: args.PARAM_CONTENT=None,
             dryrun: args.DRYRUN=False,
             wait: args.WAIT=False) -> dict:
    """Update an SSM Parameter.
    Usage: cli usage
      ```sh
      duploctl ssm_param update <name> -pval <newvalue>
      ```
    
    Args:
      name: The name of the SSM Parameter to find.
      -strat/--strategy: whether to merge or overwrite StringList Parameters (default is merge, not used for SecureString or String params)
      -pval/--parametervalue: The new value for the SSM Parameter.  Overwrites existing unless merging with StringList parameters.

    Returns: 
      resource: The SSM Parameter object.
      
    Raises:
      DuploError: If the SSM Parameter could not be found or doesn't exist.
    """
    body=self.find(name)
    if strategy=='merge' and body['Type']=="StringList":
      current_value = body['Value'].split(',')
      current_value.append(value)
      body['Value'] = ','.join(current_value)
    else:
      body['Value'] = value
    return super().update(body)



from duplocloud.client import DuploClient
from duplocloud.resource import DuploResource
from duplocloud.errors import DuploError
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("user")
class DuploUser(DuploResource):
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
    self.tenent_svc = duplo.load('tenant')

  @Command("ls")
  def list(self):
    """Retrieve a list of all users in the Duplo system."""
    response = self.duplo.get("admin/GetAllUserRoles")
    return response.json()
  
  @Command("get")
  def find(self, 
           name: args.NAME):
    """Find a User by their username."""
    try:
      return [u for u in self.list() if u["Username"] == name][0]
    except IndexError:
      raise DuploError(f"User '{name}' not found", 404)
  
  @Command()
  def create(self, 
             body: args.BODY) -> dict:
    """Create a new user.

    Usage: CLI Usage
      ```sh
      duploctl user create -f 'user.yaml'
      ```
      Contents of the `user.yaml` file
      ```yaml
      --8<-- "src/tests/data/user.yaml"
      ```
    
    Args:
      body: The user body. 

    Returns:
      message: A success message.
    """
    if 'State' not in body:
      body['State'] = 'added'
    response = self.duplo.post("admin/UpdateUserRole", body)
    return response.json()
  
  @Command()
  def delete(self, 
             name: args.NAME) -> dict:
    """Delete a User.
    
    Usage: CLI Usage
      ```sh
      duploctl user delete <name>
      ```
    
    Args:
      name: The name of the user to delete.
    
    Returns:
      message: A success message.
    """
    body = {
      "Username": name,
      "State": "deleted"
    }
    self.duplo.post("admin/UpdateUserRole", body)
    return {"message": name+ " deleted successfully"}



import pathlib
import pytest 
# import unittest
import argparse
from duplocloud.commander import schema, resources, Command, get_parser, aliased_method, extract_args, available_resources, load_resource
from duplocloud.argtype import Arg, DataMapAction
from duplocloud.errors import DuploError
# from duplo_resource.service import DuploService
# from duplocloud.resource import DuploResource

dir = pathlib.Path(__file__).parent.resolve()

NAME = Arg("name", 
            help='A test name arg')

ENABLED = Arg("enabled", "-y",
              help='A test enabled arg',
              action="store_true",
              type=bool,
              default=False)
              
IMAGE = Arg("image", "-i", "--img",
            help='A test image arg',
            default="ubuntu")

class SomeResource():
  @Command("test")
  def tester(self, 
             # use shared arg
             name: NAME,
             # inline an Arg definition with alt flag and new dest based on arg name
             image_name: IMAGE="ubuntu",
             # a boolean arg too
             enabled: ENABLED=False,
             # foo should not be registered as an arg
             foo: str="bar"):
    print(name, image_name, foo, enabled)
  def not_a_command(self):
    pass

@pytest.mark.unit
def test_command_registration():
  qn = SomeResource.tester.__qualname__
  # assert qn is a key in schema
  assert qn in schema

@pytest.mark.unit
def test_extracting_args():
  args = extract_args(SomeResource.tester)
  assert len(args) == 3
  for arg in args:
    # assert arg is an Arg
    assert isinstance(arg, Arg)
    print(arg.__name__)
    if arg.__name__ == "image":
      assert arg.default == "ubuntu"
      assert arg.attributes["dest"] == "image_name"
    if arg.__name__ == "enabled":
      assert arg.attributes["action"] == "store_true"
      assert arg.default == False # noqa: E712
      assert isinstance(arg(True), bool)

@pytest.mark.unit
def test_using_parser():
  # now make sure the proper parser is returned
  args = extract_args(SomeResource.tester)
  assert (p := get_parser(args))
  assert isinstance(p, argparse.ArgumentParser)
  # test with no args and get defaults
  args = ["foo"]
  parsed_args = p.parse_args(args)
  assert parsed_args.name == "foo"
  assert parsed_args.image_name == "ubuntu"
  # test with image set
  args = ["bar", "--img", "alpine"]
  parsed_args = p.parse_args(args)
  assert parsed_args.name == "bar"
  assert parsed_args.image_name == "alpine"
  # one more time with --image
  args = ["baz", "--image", "splunz:latest"]
  parsed_args = p.parse_args(args)
  assert parsed_args.name == "baz"
  assert parsed_args.image_name == "splunz:latest"

@pytest.mark.unit
def test_loading_service():
  assert (svc := load_resource("service")) # noqa: F841
  assert "service" in resources
  svcs = available_resources()
  assert "service" in svcs

@pytest.mark.unit
def test_arg_type():
  assert isinstance(NAME, Arg)
  name = NAME("foo")
  assert isinstance(name, str)

@pytest.mark.unit
def test_aliased_command():
  method = aliased_method(SomeResource, "test")
  assert method == "tester"

@pytest.mark.unit
def test_datamap_action():
  fname = "password.txt"
  password = "verysecretpassword"
  fpath = f"{dir}/files/{fname}"
  p = argparse.ArgumentParser()
  p.add_argument("--from-file", "--from-literal", dest="data", action=DataMapAction)
  args = [
    "--from-file", fpath,
    "--from-file", f"renamed={fpath}",
    "--from-literal", "foo=bar"
  ]
  ns = p.parse_args(args)
  assert ns.data == {
    "password.txt": password,
    "renamed": password,
    "foo": "bar"
  }



from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("cronjob")
class DuploCronJob(DuploTenantResourceV3):
  """
  Duplo CronJob are scheduled jobs that run containers in a Kubernetes cluster.
  """

  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, "k8s/cronJob")

  @Command()
  def update_image(self,
                   name: args.NAME,
                   image: args.IMAGE):
    """Update the image of a cronjob.

    Args:
      name (str): The name of the cronjob to update.
      image (str): The new image to use for the cronjob.
    """
    cronjob = self.find(name)

    # Set IsAnyHostAllowed
    self._set_is_any_host_allowed(cronjob)

    cronjob["spec"]["jobTemplate"]["spec"]["template"]["spec"]["containers"][0]["image"] = image
    self.update(cronjob)
    return {"message": f"Successfully updated image for cronjob '{name}'"}

  @Command()
  def update_schedule(self,
                      name: args.NAME,
                      cronschedule: args.CRONSCHEDULE) -> dict:
    """Update the schedule of a cronjob.

    Args:
      name: The name of the cronjob to update.
      cronschedule: The new schedule to use for the cronjob.
    Returns:
      message: A success message.
    """
    cronjob = self.find(name)

    # Set IsAnyHostAllowed
    self._set_is_any_host_allowed(cronjob)

    cronjob["spec"]["schedule"] = cronschedule
    self.update(cronjob)
    return {"message": f"Successfully updated cron-schedule for cronjob '{name}'"}

  def _set_is_any_host_allowed(self, cronjob: dict) -> None:
    """Helper method to set the 'IsAnyHostAllowed' field based on cronjob annotations."""
    # Check if the annotation "duplocloud.net/is-any-host-allowed" exists and set `IsAnyHostAllowed`
    annotations = cronjob.get("metadata", {}).get("annotations", {})
    is_any_host_allowed = annotations.get("duplocloud.net/is-any-host-allowed")

    if is_any_host_allowed == "true":
      cronjob["IsAnyHostAllowed"] = True
    else:
      cronjob["IsAnyHostAllowed"] = False


from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("cloudfront")
class CloudFront(DuploTenantResourceV3):

    def __init__(self, duplo: DuploClient):
        super().__init__(duplo, "aws/cloudfront")
        self.wait_timeout = 1200

    @Command()
    def create(self, body: args.BODY):
        """Create a CloudFront distribution.
        
        Args:
          body (dict): The body of the request.
        """
        response = self.duplo.post(self.endpoint, body)
        return response.json()

    @Command()
    def list(self):
        """List CloudFront distributions."""
        response = self.duplo.get(self.endpoint)
        return response.json()

    @Command()
    def disable(self, distribution_id: args.NAME):
        """Disable a CloudFront distribution.
        
        Args:
          distribution_id (str): The ID of the distribution to disable.
        """
        body = {"Status": "Disabled"}
        response = self.duplo.put(self.endpoint(distribution_id, "update"), body)
        return response.json()

    @Command()
    def delete(self, distribution_id: args.NAME):
        """Delete a CloudFront distribution.
        
        Args:
          distribution_id (str): The ID of the distribution to delete.
        """
        self.duplo.delete(self.endpoint(distribution_id))
        return {"message": f"CloudFront distribution {distribution_id} deleted"}



import argparse
from copy import deepcopy
from importlib.metadata import entry_points, version
from inspect import signature, getmro, Parameter
from .errors import DuploError
from .argtype import Arg
from typing import Callable, List, Type

ENTRYPOINT="duplocloud.net"
FORMATS=f"formats.{ENTRYPOINT}"
VERSION=version('duplocloud-client')
ep = entry_points(group=ENTRYPOINT)
fep = entry_points(group=FORMATS)
schema = {}
resources = {}

def Resource(name: str):
  def decorator(cls):
    resources[name] = {
      "class": cls.__qualname__
    }
    setattr(cls, "kind", name)
    return cls
  return decorator

def Command(*aliases) -> Callable:
  """Command decorator

  This decorator is used to register a function as a command. It will
  automatically generate the command line arguments for the function
  based on the annotations.

  Example:
    ```python
    from duplocloud.commander import Command
    from duplocloud import args
    @Command()
    def hello(name: args.NAME = "world"):
      print(f"Hello {name}!")
    ```
  
  Returns:
    decorator: The decorated function.

  """
  def decorator(function: Callable):
    schema[function.__qualname__] = {
      "class": function.__qualname__.split(".")[0],
      "method": function.__name__,
      "aliases": list(aliases),
    }
    return function
  return decorator

def extract_args(function: Callable) -> List[Arg]:
  """Extract Args
  
  Extract the cli argument annotations from a function. This will only collect the args of type duplocloud.Arg. 
  This list can now be used to generate an argparse.ArgumentParser object.
  
  """
  sig = signature(function)
  def arg_anno(name, param):
    a = deepcopy(param.annotation)
    if not a.positional and name != a.__name__:
      a.set_attribute("dest", name)
    # hmmm, is this a good idea? If so, maybe some innovations is needed. 
    # if param.default is not Parameter.empty:
    #   a.set_attribute("default", param.default)
    return a
  return [
    arg_anno(k, v)
    for k, v in sig.parameters.items()
    if v.annotation is not Parameter.empty and isinstance(v.annotation, Arg)
  ]

def aliased_method(cls: Type, command: str) -> str:
  """Aliased Method
  
  Given a name of a command, check the schema and find the real method name because the command might be aliased.
  The given class will be used to discover any ancestors because the command may actually come from a parent class.

  Args:
    cls: The class to check the schema for.
    command: The command to find the

  Returns:
    method: The true name of the commands method.
  """
  clss = [c.__name__ for c in getmro(cls) if c.__name__ != "object"]
  s = next((
    v for v in schema.values()
    if (v["class"] in clss) and (command == v["method"] or command in v.get("aliases", []))
  ), None)
  if not s:
    raise DuploError(f"Command {command} not found.", 404)
  return s["method"]

def get_parser(args: List[Arg]) -> argparse.ArgumentParser:
  """Get Parser
  
  Args:
    args: A list of Arg objects.
  Returns:
    An argparse.ArgumentParser object with args from function.
  """
  parser = argparse.ArgumentParser(
    prog='duplocloud-client',
    description='Duplo Cloud CLI',
  )
  for arg in args:
    parser.add_argument(*arg.flags, default=arg.default, **arg.attributes)
  return parser

def load_resource(name: str):
  """Load Service
    
  Load a Service class from the entry points.

  Args:
    name: The name of the service.
  Returns:
    The class of the service.
  """
  try:
    return ep[name].load()
  except KeyError:
    avail = available_resources()
    raise DuploError(f"""
Resource named {name} not found.
Available resources are:
  {", ".join(avail)}
""", 404)

def load_format(name: str="string"):
  """Load Format
    
  Load a Formatter function from the entry points.

  Args:
    name: The name of the format.
  Returns:
    The class of the format.
  """
  return fep[name].load()

def available_resources() -> List[str]:
  """Available Resources

  Returns:
    A list of available resources names.
  """
  return list(ep.names)

def available_formats() -> List[str]:
  """Available Formats

  Returns:
    A list of available format names.
  """
  return list(fep.names)



import pytest
from duplocloud.client import DuploClient
from duplocloud.errors import DuploError

class TestJIT:
  
  @pytest.mark.integration
  @pytest.mark.order(10)
  def test_admin_aws_jit(self, duplo: DuploClient):
    duplo.isadmin = True
    r = duplo.load("jit")
    try:
      o = r.aws()
      assert o.get("SecretAccessKey", None) is not None
    except DuploError as e:
      pytest.fail(f"Failed to get admin credentials: {e}")
  
  @pytest.mark.integration
  @pytest.mark.order(10)
  def test_user_aws_jit(self, duplo: DuploClient):
    duplo.isadmin = False
    r = duplo.load("jit")
    try:
      o = r.aws()
      assert o.get("SecretAccessKey", None) is not None
    except DuploError as e:
      pytest.fail(f"Failed to get user credentials: {e}")
  
  @pytest.mark.integration
  @pytest.mark.order(10)
  def test_user_k8s_context(self, duplo: DuploClient):
    duplo.isadmin = False
    r = duplo.load("jit")
    try:
      o = r.k8s_context()
      assert o.get("Token", None) is not None
    except DuploError as e:
      pytest.fail(f"Failed to get user k8s credentials: {e}")
  
  @pytest.mark.integration
  @pytest.mark.order(10)
  def test_admin_k8s_context(self, duplo: DuploClient):
    duplo.isadmin = True
    r = duplo.load("jit")
    try:
      o = r.k8s_context()
      assert o.get("Token", None) is not None
    except DuploError as e:
      pytest.fail(f"Failed to get admin k8s credentials: {e}")
  
  




from duplocloud.client import DuploClient
from duplocloud.commander import Resource
from importlib.metadata import version

@Resource("version")
class DuploVersion():
  """Show Version

  Prints the version of the Duplo CLI, UI, and Server.

  Usage:
    ```sh
    duploctl version
    ```
  
  Returns:
    version (dict): The Duplo version.
  """
  def __init__(self, duplo: DuploClient):
    self.duplo = duplo
    self.paths = {
      "ui": "build-metadata.json", 
      "frontdoor": "frontdoor/build-metadata.json",
      "backend": "v3/version",
      "auth": "v3/auth/version",
      "katkit": "v3/katkit/version",
      "billing": "v3/billing/version",
      "security": "v3/security/version"
    }
    
  def __call__(self) -> dict:
    v = {}
    # first get cli version then server versions
    v["cli"] = {
      "tag": version('duplocloud-client')
    }
    for (name, path) in self.paths.items():
      try:
        res = self.duplo.get(path).json()
      except Exception as e:
        self.duplo.logger.error(f"Failed to get version for {path}: {e}")
        res = {"error": str(e)}
      finally:
        v[name] = res
    return v



from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV2
from duplocloud.errors import DuploError
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("asg")
class DuploAsg(DuploTenantResourceV2):
  """Manage Duplo ASGs

  Duplo ASGs are Auto Scaling Groups that manage the number of hosts within a tenant.
  """
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
  
  @Command()
  def list(self) -> list:
    """List all ASGs
    
    Usage: CLI Usage
      ```sh
      duploctl asg list
      ```
    Returns:
      list: A list of all ASGs.
    """
    tenant_id = self.tenant["TenantId"]
    response = self.duplo.get(f"subscriptions/{tenant_id}/GetTenantAsgProfiles")
    return response.json()

  @Command()
  def find(self, 
           name: args.NAME):
    """Find an asg by name.

    Usage: CLI Usage
      ```sh
      duploctl asg find <name>
      ```
    Args:
      name (str): The name of the asg to find.
    Returns: 
      The service object.
    Raises:
      DuploError: If the asg could not be found.
    """
    try:
      return [s for s in self.list() if s["FriendlyName"] == name][0]
    except IndexError:
      raise DuploError(f"ASG Profile '{name}' not found", 404)
    
  @Command()
  def create(self,
             body: args.BODY,
             wait: args.WAIT=False) -> dict:
    """Create an ASG
    
    Usage: CLI Usage
      ```sh
      duploctl hosts create -f 'asg.yaml'
      ```
      Contents of the `asg.yaml` file
      ```yaml
      --8<-- "src/tests/data/asg.yaml"
      ```
    Example: One liner example
      ```sh
      echo \"\"\"
      --8<-- "src/tests/data/asg.yaml"
      \"\"\" | duploctl asg create -f -
      ```
    Args:
      body: The body of the request.
      wait: Whether to wait for the creation to complete.
    Returns:
      message: A message that the asg was successfully created.
    """
    tenant_id = self.tenant["TenantId"]
    name = self.name_from_body(body)
    if body.get("ImageId", None) is None:
      body["ImageId"] = self.discover_image(body.get("AgentPlatform", 0))
    res = self.duplo.post(f"subscriptions/{tenant_id}/UpdateTenantAsgProfile", body)
    def wait_check():
      return self.find(name)
    if wait:
      self.wait(wait_check)
    return {
      "message": f"Successfully created asg '{body['FriendlyName']}'",
      "data": res.json()
    }
  
  @Command()
  def update(self,
             body: args.BODY) -> dict:
    """Update an ASG.

    Usage: CLI Usage
      ```sh
      duploctl asg update -f <file>
      ```
    
    Args:
      body: The body of the request.

    Returns:
      message: A message that the asg was successfully updated
    """
    tenant_id = self.tenant["TenantId"]
    res = self.duplo.post(f"subscriptions/{tenant_id}/UpdateTenantAsgProfile", body)
    return {
      "message": f"Successfully updated asg '{body['FriendlyName']}'",
      "data": res.json()
    }
  
  @Command()
  def delete(self,
             name: args.NAME) -> dict:
    """Delete an ASG

    Usage: CLI Usage
      ```sh
      duploctl asg delete <name>
      ```
    
    Args:
      name: The name of the asg to delete.

    Returns:
      message: A message that the asg was successfully deleted.
    """
    tenant_id = self.tenant["TenantId"]
    body = { 
      "FriendlyName": name,
      "State": "delete"
    }
    res = self.duplo.post(f"subscriptions/{tenant_id}/UpdateTenantAsgProfile", body)
    return {
      "message": f"Successfully deleted asg '{name}'",
      "data": res.json()
    }

  @Command()
  def scale(self,
            name: args.NAME,
            min: args.MIN=None,
            max: args.MAX=None) -> dict:
    """Scale an ASG.
    
    Note: "-m" represents Minimum instances and "-M" represents Maximum instances.

    Usage: CLI Usage
      ```sh
      duploctl asg scale -n <name> [-m <min>] [-M <max>]
    
    Args:
      name: The name of the asg to scale.
      min: The minimum number of instances.
      max: The maximum number of instances.

    Returns:
      message: A messaget that the asg was successfully scaled.
    """
    if not min and not max:
      raise DuploError("Must provide either min or max")
    asg = self.find(name)
    data = {
      "FriendlyName": name,
      "DesiredCapacity": asg.get("MinSize", None),
      "MinSize": asg.get("MinSize", None),# this really is a string unlike the other two? 
      "MaxSize": asg.get("MaxSize", None),
    }
    if min:
      data["MinSize"] = str(min)
    if max:
      data["MaxSize"] = max
    return self.update(data)
  
  def name_from_body(self, body):
    prefix = f"duploservices-{self.tenant['AccountName']}"
    name =  body["FriendlyName"]
    if not name.startswith(prefix):
      name = f"{prefix}-{name}"
    return name

  def discover_image(self, agent, arch="amd64"):
    imgs = self.tenant_svc.host_images(self.tenant['AccountName'])
    try:
      img = [i for i in imgs if i["Agent"] == agent and i["Arch"] == arch][0]
      return img.get("ImageId")
    except IndexError:
      raise DuploError(f"Image for agent '{agent}' not found", 404)



from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV2
from duplocloud.errors import DuploError
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("lambda")
class DuploLambda(DuploTenantResourceV2):
  """Manage Duplo Lambdas
  
  Duplo Lambdas are serverless functions that run in response to events.
  """
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
  
  @Command()
  def list(self) -> list:
    """List Lambdas.

    List all of the tenants in the current tenant.

    Usage: CLI Usage
      ```sh
      duploctl lambda list
      ```
    
    Returns:
      list: A list of all lambdas in the current subscription.
    """
    tenant_id = self.tenant["TenantId"]
    tenant_name = self.tenant["AccountName"]
    response = self.duplo.get(f"subscriptions/{tenant_id}/GetLambdaFunctions")
    if (data := response.json()):
      return data
    else:
      raise DuploError(f"No lambda functions found in tenant '{tenant_name}'", 404)
  
  @Command()
  def find(self, 
           name: args.NAME) -> dict:
    """Find a Lambda function by name.

    Usage: CLI Usage
      ```sh
      duploctl lambda find <name>
      ```
    Args:
      name: The name of the lambda to find.
    Returns: 
      The lambda object.
    Raises:
      DuploError: If the lambda could not be found.
    """
    try:
      return [s for s in self.list() if s["FunctionName"] == name][0]
    except IndexError:
      raise DuploError(f"Lambda '{name}' not found", 404)
    
  @Command()
  def create(self, 
             body: args.BODY,
             wait: args.WAIT = False) -> dict:
    """Create a new tenant.

    Usage: CLI Usage
      ```sh
      duploctl lambda create -f 'lambda.yaml'
      ```
      Contents of the `lambda.yaml` file
      ```yaml
      --8<-- "src/tests/data/lambda.yaml"
      ```
    
    Args:
      body: The lambda body. 
      wait: Whether to wait for the tenant to be created

    Returns:
      message: A success message.
    """
    def wait_check():
      name = self.name_from_body(body)
      self.find(name)
    tenant_id = self.tenant["TenantId"]
    self.duplo.post(f"subscriptions/{tenant_id}/CreateLambdaFunction", body)
    if wait:
      self.wait(wait_check, 400)
    return {
      "message": f"Lambda {body['FunctionName']} created"
    }
  
  @Command()
  def delete(self, 
             name: args.NAME) -> dict:
    """Delete a lambda.

    Usage: CLI Usage
      ```sh
      duploctl lambda delete <name>
      ```
    
    Args:
      name: The name of the lambda to delete.
    
    Returns:
      message: A success message.
    """
    tenant_id = self.tenant["TenantId"]
    self.duplo.post(f"subscriptions/{tenant_id}/DeleteLambdaFunction/{name}")
    return {
      "message": f"Lambda {name} deleted"
    }

  @Command()
  def update_image(self, 
                   name: args.NAME, 
                   image: args.IMAGE) -> dict:
    """Update the image of a lambda.

    Usage: CLI Usage
      ```sh
      duploctl lambda update_image <name> <image>
      ```
    
    Args:
      name: The name of the lambda to update.
      image: The new image to use for the lambda.
    
    Returns:
      message: The updated lambda object
    """
    tenant_id = self.tenant["TenantId"]
    data = {
      "FunctionName": name,
      "ImageUri": image
    }
    response = self.duplo.post(f"subscriptions/{tenant_id}/UpdateLambdaFunction", data)
    return response.json()

  @Command()
  def update_s3(self, 
                name: args.NAME, 
                bucket: args.S3BUCKET,
                key: args.S3KEY) -> dict:
    """Update the s3 bucket and key of a lambda.

    Usage: CLI Usage
      ```sh
      duploctl lambda update_s3 <name> <bucket> <key>
      ```
    
    Args:
      name: The name of the lambda to update.
      bucket: The s3 bucket to use for the lambda.
      key: The s3 key (file path) to use for the lambda.
    """
    tenant_id = self.tenant["TenantId"]
    data = {
      "FunctionName": name,
      "S3Bucket": bucket,
      "S3Key": key
    }
    response = self.duplo.post(f"subscriptions/{tenant_id}/UpdateLambdaFunction", data)
    return response.json()

  def name_from_body(self, body):
    prefix = f"duploservices-{self.tenant['AccountName']}"
    name =  body["FunctionName"]
    if not name.startswith(prefix):
      name = f"{prefix}-{name}"
    return name

  @Command()
  def update_env(self,
               name: args.NAME,
               setvar: args.SETVAR,
               strategy: args.STRATEGY,
               deletevar: args.DELETEVAR) -> dict:
    """Update the environment variables of a lambda. If lambda has no environment variables set, use -strat replace to set new values.

    Usage: Basic CLI Use
      ```sh
      duploctl lambda update_env <lambda-name> --setvar env-key1 env-val1 --setvar env-key2 env-val2 --setvar env-key3 env-val3 -strat merge
      ```

    Args:
      name (str): The name of the lambda to update.
      setvar/-V (list): A list of key-value pairs to set as environment variables.
      strategy/strat (str): The merge strategy to use for env vars. Valid options are "merge" or "replace". Default is merge.
      deletevar/-D (list): A list of keys to delete from the environment variables.
    """
    tenant_id = self.tenant["TenantId"]
    current_lambda = self.find(name)
    current_env = current_lambda.get('Environment', {}).get('Variables', {})
    new_env = {k: v for k, v in (setvar or [])}

    if strategy == 'merge':
      merged_env = {**current_env, **new_env}
    else:
      merged_env = new_env

    if deletevar:
      for key in deletevar:
        merged_env.pop(key, None)

    payload = {
      "FunctionName": name,
      "Environment": {"Variables": merged_env}
    }

    response = self.duplo.post(f"subscriptions/{tenant_id}/UpdateLambdaFunctionConfiguration", payload)
    if response.ok:
      return {"message": f"Successfully updated environment variables for Lambda '{name}'"}
    else:
      return {
        "message": f"Failed to update environment variables for Lambda '{name}'",
        "status_code": response.status_code,
        "error": response.text
      }



import math
import pytest
import random

from duplocloud.client import DuploClient
from duplocloud.errors import DuploError, DuploFailedResource

class TestInfra:

  # def setup_class(self):
  #   inc = random.randint(1, 100)
  #   self.infra_name = f"duploctl{inc}"

  @pytest.mark.integration
  def test_listing_infrastructures(self, duplo: DuploClient):
    r = duplo.load("infrastructure")
    try:
      lot = r("list")
    except DuploError as e:
      pytest.fail(f"Failed to list infrastructures: {e}")

  @pytest.mark.integration
  def test_finding_infra(self, duplo: DuploClient):
    r = duplo.load("tenant")
    try:
      t = r("find", "default")
    except DuploError as e:
      pytest.fail(f"Failed to find default infra: {e}")
    assert t["AccountName"] == "default"

  @pytest.mark.integration
  @pytest.mark.dependency(name = "create_infra", scope='session')
  @pytest.mark.order(1)
  def test_creating_infrastructures(self, duplo: DuploClient, infra_name: str, e2e: bool):
    r = duplo.load("infrastructure")
    vnum = math.ceil(random.randint(1, 9))
    if e2e:
      duplo.tenant = infra_name
    # check if the infra already exists
    try:
      i = r.find(infra_name)
      if i:
        pytest.skip(f"Infrastructure '{infra_name}' already exists")
    except DuploError as e:
      pass
    name = infra_name
    print(f"Creating infra '{name}'")
    try:
      r.create({
        "Name": name,
        "Accountid": "",
        "EnableK8Cluster": True,
        "AzCount": 2,
        "Vnet": { 
          "SubnetCidr": 22, 
          "AddressPrefix": f"11.1{vnum}0.0.0/16"
        },
        "Cloud": 0,
        "OnPremConfig": None,
        "Region": "us-east-2",
        "CustomData": [],
      }, wait=True)
    except DuploFailedResource as e:
      pytest.fail(f"Infrastructure is in a failed state: {e}")
    except DuploError as e:
      pytest.fail(f"Failed to create tenant: {e}")
    
  @pytest.mark.integration
  @pytest.mark.dependency(depends=["create_infra"], scope='session')
  @pytest.mark.order(999)
  def test_find_delete_infra(self, duplo: DuploClient, infra_name: str):
    r = duplo.load("infrastructure")
    # name = self.infra_name
    name = infra_name
    print(f"Deleting infra '{name}'")
    try:
      i = r.find(name)
      assert i["Name"] == name
    except DuploError as e:
      pytest.fail(f"Failed to find infrastructure {name}: {e}")
    try:
      r("delete", name)
    except DuploError as e:
      pytest.fail(f"Failed to delete infrastructure: {e}")



from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Resource


@Resource("storageclass")
class DuploStorageClass(DuploTenantResourceV3):
  """
  DuploStorageClass is a resource that represents a Kubernetes StorageClass.  

  See  
   - https://docs.duplocloud.com/docs/overview/aws-services/storage/adding-k8s-storage-class
  """
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, "k8s/storageclass")

  def name_from_body(self, body):
    return body["name"]



from duplocloud import args
from duplocloud.client import DuploClient
from duplocloud.errors import DuploError
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Command, Resource

@Resource("aws_secret")
class DuploAwsSecret(DuploTenantResourceV3):
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, "aws/secret")
  
  def name_from_body(self, body):
    return body["Name"]
  
  @Command()
  def create(self, 
             name: args.NAME=None,
             body: args.BODY=None,
             value: args.PARAM_CONTENT=None,
             dryrun: args.DRYRUN=False,
             wait: args.WAIT=False) -> dict:
    """Create an AWS Secretmanager Secret
    Usage: cli usage
      ```sh
      duploctl aws_secret create <name> -pval <value>
      ```
    
    Args:
      name: The name of the AWS Secret to create.
      -pval/--parametervalue: Arbitrary text to set in the AWS secret.
      -body: path to a raw json/yaml post body, e.g:
      ```
      {
        Name: "test2", 
        SecretString: '{"Foo": "Bar", "Biz":"Baz"}'
      }
      ```

    Returns: 
      resource: The AWS secret object.
      
    Raises:
      DuploError: If the AWS secret already exists.
    """
    # Make sure the user passes name and value, or a body (from file).
    if not body and (not name or not value):
      raise DuploError("name and value are required when body is not provided")
    if not body:
      body = {}
    # If the user passed in settings, use them.  Otherwise use what's in the file.
    if name:
      body['Name'] = name
    if value:
      body['SecretString'] = value
    if dryrun:
      return body
    else:
      return super().create(body, wait=wait)

  @Command()
  #Implement find with opt-in option to display sensitive data. 
  #Note that using find still returns the sensitive data!
  def find(self, 
           name: args.NAME,
           show_sensitive: args.SHOW_SENSITIVE=False) -> dict:
    """Find as AWS Secretmanager secret by name and return it's content

    Usage: cli usage
      ```sh
      duploctl aws_secret find <name>
      ```
    
    Args:
      name: The name of the AWS secret to find.
      -show/--showsensitive: Display value of the secretstring field

    Returns: 
      resource: The AWS secret object.
      
    Raises:
      DuploError: If the AWS secret could not be found.
    """
    response = self.duplo.get(self.endpoint(name))
    if not show_sensitive:
      obfuscated_response=response.json()
      sensitive_len=len(response.json()["SecretString"])
      placeholder="*"
      obfuscated_response["SecretString"]=placeholder * sensitive_len
      return obfuscated_response
    else:
      return response.json()
    
  @Command()
  def update(self, 
             name: args.NAME=None,
             value: args.PARAM_CONTENT=None,
             dryrun: args.DRYRUN=False,
             wait: args.WAIT=False) -> dict:
    """Update an AWS Secretmanager secret.
    Usage: cli usage
      ```sh
      duploctl aws_secret update <name> -pval <newvalue>
      ```
    
    Args:
      name: The name of the AWS secret to find.
      -pval/--parametervalue: The new value for the AWS secret.  This overwrites the existing value!

    Returns: 
      resource: The AWS secret object.
      
    Raises:
      DuploError: If the AWS secret could not be found or doesn't exist.
    """
    body=self.find(name)
    body['SecretString'] = value
    return super().update(body)



from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV2
from duplocloud.errors import DuploError, DuploFailedResource
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("hosts")
class DuploHosts(DuploTenantResourceV2):
  """Manage Duplo Hosts
  
  Duplo hosts are virtual machines that run your services. You can create, delete, start, stop, and reboot hosts.
  """
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
    self.paths = {
      "list": "GetNativeHosts"
    }
  
  @Command()
  def create(self,
             body: args.BODY,
             wait: args.WAIT=False) -> dict:
    """Create a Hosts resource.

    Usage: CLI Usage
      ```sh
      duploctl hosts create -f 'hosts.yaml'
      ```
      Contents of the `hosts.yaml` file
      ```yaml
      --8<-- "src/tests/data/hosts.yaml"
      ```

    Example: One liner example
      ```sh
      echo \"\"\"
      --8<-- "src/tests/data/hosts.yaml"
      \"\"\" | duploctl hosts create -f -
      ```
    
    Args:
      body: The resource to create.
      wait: Wait for the resource to be created.

    Returns: 
      message: Success message.

    Raises:
      DuploError: If the resource could not be created.
    """
    def wait_check():
      name = self.name_from_body(body)
      h = self.find(name)
      if h["Status"] != "running":
        if h["Status"] != "pending":
          raise DuploFailedResource(f"Host '{name}' failed to create.")
        raise DuploError(f"Host '{name}' not ready", 404)
    # let's get started
    if body.get("ImageId", None) is None:
      body["ImageId"] = self.discover_image(body.get("AgentPlatform", 0))
    res = self.duplo.post(self.endpoint("CreateNativeHost"), body)
    if wait:
      self.wait(wait_check)
    return {
      "message": f"Successfully created host '{body['FriendlyName']}'",
      "id": res.json()
    }
  
  @Command()
  def delete(self,
             name: args.NAME,
             wait: args.WAIT=False) -> dict:
    """Delete a host.
    
    Deletes a host by name. If the host is running, it will be stopped before deletion.

    Usage: cli
      ```sh
      duploctl hosts delete <name>
      ```
    
    Args:
      name: The name of the host to delete.
      wait: Wait for the host to be deleted.

    Returns:
      message: A success message.
    """
    host = self.find(name)
    inst_id = host["InstanceId"]
    res = self.duplo.post(self.endpoint(f"TerminateNativeHost/{inst_id}"), host)
    def wait_check():
      h = None 
      try:
        h = self.find(name)
      except DuploError as e:
        if e.code == 404:
          return None # if 404 then it's gone so finish waiting
        else:
          raise DuploFailedResource(f"Host '{name}' failed to delete.")
      if h["Status"] == "shutting-down" or h["Status"] == "running":
        raise DuploError(f"Host '{name}' not terminated", 404)
    if wait:
      self.wait(wait_check, 500)
    return {
      "message": f"Successfully deleted host '{name}'",
      "data": res.json()
    }
  
  @Command()
  def stop(self,
           name: args.NAME,
           wait: args.WAIT=False) -> dict:
    """Stop a host.
    
    Stops a host by name. If the host is already stopped, it will return a success message.

    Usage: cli
      ```sh
      duploctl hosts stop <name>
      ```
    
    Args:
      name: The name of the host to stop.
      wait: Wait for the host to stop.

    Returns:
      message: A success message.
    """
    host = self.find(name)
    inst_id = host["InstanceId"]
    res = self.duplo.post(self.endpoint(f"stopNativeHost/{inst_id}"), host)
    def wait_check():
      h = self.find(name)
      if h["Status"] == "running":
        raise DuploError(f"Host '{name}' not ready", 404)
      if h["Status"] != "stopped":
        if h["Status"] != "stopping":
          raise DuploFailedResource(f"Host '{name}' failed to stop.")
        raise DuploError(f"Host '{name}' not ready", 404)
    if wait:
      self.wait(wait_check, 500)
    return {
      "message": f"Successfully stopped host '{name}'",
      "data": res.json()
    }
  
  @Command()
  def start(self,
             name: args.NAME,
             wait: args.WAIT=False) -> dict:
    """Start a host.
    
    Starts a host by name. If the host is already running, it will return a success message.

    Usage: cli
      ```sh
      duploctl hosts start <name>
      ```
    
    Args:
      name: The name of the host to start.
      wait: Wait for the host to start.
    
    Returns:
      message: A success message.
    """
    host = self.find(name)
    inst_id = host["InstanceId"]
    res = self.duplo.post(self.endpoint(f"startNativeHost/{inst_id}"), host)
    def wait_check():
      h = self.find(name)
      if h["Status"] == "stopped":
        raise DuploError(f"Host '{name}' not ready", 404)
      if h["Status"] != "running":
        if h["Status"] != "pending":
          raise DuploFailedResource(f"Host '{name}' failed to stop.")
        raise DuploError(f"Host '{name}' not ready", 404)
    if wait:
      self.wait(wait_check, 500)
    return {
      "message": f"Successfully started host '{name}'",
      "data": res.json()
    }
  
  @Command()
  def reboot(self,
             name: args.NAME) -> dict:
    """Reboot a host
    
    Reboots a host by name.

    Usage: cli
      ```sh
      duploctl hosts reboot <name>
      ```

    Args:
      name: The name of the host to reboot.
    
    Returns:
      message: A success message.
    """
    host = self.find(name)
    inst_id = host["InstanceId"]
    res = self.duplo.post(self.endpoint(f"RebootNativeHost/{inst_id}"), host)
    return {
      "message": f"Successfully rebooted host '{name}'",
      "data": res.json()
    }
    
  def name_from_body(self, body):
    prefix = f"duploservices-{self.tenant['AccountName']}"
    name =  body["FriendlyName"]
    if not name.startswith(prefix):
      name = f"{prefix}-{name}"
    return name
  
  def discover_image(self, agent, arch="amd64"):
    imgs = self.tenant_svc.host_images(self.tenant['AccountName'])
    try:
      img = [i for i in imgs if i["Agent"] == agent and i["Arch"] == arch][0]
      return img.get("ImageId")
    except IndexError:
      raise DuploError(f"Image for agent '{agent}' not found", 404)
